from __future__ import annotations

import math
from collections import Counter
from datetime import datetime, timedelta
from typing import Any

import orjson
import polars as pl

from uforia.storage import DatasetName, DatasetStore
from uforia.storage.datasets import DATASET_SPECS, ReadMode, empty_frame

SOURCE = "perps_family_v1"
COMPOSITE_SOURCE = "perps_composite_v1"
TRANSITION_SOURCE = "perps_transition_v1"
SECONDS_PER_YEAR = 365.0 * 24.0 * 60.0 * 60.0

REGIME_FAMILIES: tuple[tuple[str, str], ...] = (
    ("cascade", "cascade_regime"),
    ("toxicity", "toxicity_regime"),
    ("leverage", "leverage_regime"),
    ("spread", "spread_regime"),
    ("rv", "rv_regime"),
    ("delta", "delta_regime"),
)


def build_perps_signal_series(
    store: DatasetStore,
    start: datetime,
    end: datetime,
    *,
    asset: str,
    venue_scope: str = "all",
    source: str = SOURCE,
) -> pl.DataFrame:
    # Perps dashboards are meant to refresh from the hot live microstructure window.
    # A longer 5s read is too expensive for the current Mac Mini deployment and
    # most metrics degrade gracefully when older history is unavailable.
    lookback_start = start - timedelta(hours=6)
    bars_1m = _with_bar_context(
        _read_hot_window(
            store,
            DatasetName.MICROSTRUCTURE_BAR,
            lookback_start,
            end,
            filters={"asset": asset, "bar_size_secs": 60},
        )
    )
    bars_5s = _with_bar_context(
        _read_hot_window(
            store,
            DatasetName.MICROSTRUCTURE_BAR,
            lookback_start,
            end,
            filters={"asset": asset, "bar_size_secs": 5},
        )
    )
    raw_events = _with_raw_context(
        _read_hot_window(
            store,
            DatasetName.MICROSTRUCTURE_RAW_EVENT,
            lookback_start,
            end,
            filters={"asset": asset},
        )
    )
    book = _with_book_context(
        _read_hot_window(
            store,
            DatasetName.MICROSTRUCTURE_BOOK_STATE,
            lookback_start,
            end,
            filters={"asset": asset},
        )
    )
    funding = _with_funding_context(
        _read_hot_window(
            store,
            DatasetName.FUNDING_RATE_SNAPSHOT,
            lookback_start,
            end,
            filters={"underlying": asset},
        )
    )
    if bars_1m.is_empty():
        return pl.DataFrame()

    bars_1m = _filter_scope(bars_1m, venue_scope)
    bars_5s = _filter_scope(bars_5s, venue_scope)
    raw_events = _filter_scope(raw_events, venue_scope)
    book = _filter_scope(book, venue_scope)
    funding_scope = _filter_funding_scope(funding, venue_scope)

    asset_bars = _aggregate_asset_bars(bars_1m)
    if asset_bars.is_empty():
        return pl.DataFrame()

    trades = _aggregate_trade_context(raw_events)
    liquidations = _aggregate_liquidation_context(raw_events)
    open_interest = _aggregate_open_interest_context(raw_events)
    funding_context = _aggregate_funding_context(funding_scope)
    cross_venue = _aggregate_cross_venue_context(bars_1m, funding)
    vol_context = _aggregate_volatility_context(bars_5s, bars_1m)
    depth_context = _aggregate_depth_context(book, raw_events)

    base = asset_bars.join(trades, on="ts", how="left")
    for frame in (
        liquidations,
        open_interest,
        funding_context,
        cross_venue,
        vol_context,
        depth_context,
    ):
        if not frame.is_empty():
            join_columns = [
                "ts",
                *[
                    column
                    for column in frame.columns
                    if column != "ts" and column not in base.columns
                ],
            ]
            base = base.join(frame.select(join_columns), on="ts", how="left")

    missing_defaults: dict[str, Any] = {
        "long_liq_notional": None,
        "short_liq_notional": None,
        "total_liq_notional": None,
        "oi_value": None,
        "funding_rate_1h": None,
        "cross_venue_funding_spread": None,
        "venue_dislocation_bps": None,
        "binance_ret": None,
        "bybit_ret": None,
        "rv_5s": None,
        "rv_1m": None,
        "rv_5m": None,
        "depth_delta": None,
        "aggressive_flow_notional": None,
        "trade_notional": None,
        "large_trade_notional": None,
        "aggressive_buy_notional": None,
        "aggressive_sell_notional": None,
    }
    absent = [column for column in missing_defaults if column not in base.columns]
    if absent:
        base = base.with_columns(
            [pl.lit(missing_defaults[column], dtype=pl.Float64).alias(column) for column in absent]
        )

    base = base.sort("ts").with_columns(
        [
            pl.lit(asset).alias("asset"),
            pl.lit(venue_scope).alias("venue_scope"),
            pl.lit(source).alias("source"),
            (((pl.col("perp_mid") - pl.col("spot_mid")) / pl.col("spot_mid")) * 10_000.0).alias(
                "basis_bps"
            ),
            (
                (((pl.col("perp_mid") - pl.col("spot_mid")) / pl.col("spot_mid")) * 100.0)
                + (pl.col("funding_rate_1h") * 24.0 * 365.0 * 100.0)
            ).alias("carry_annualized_pct"),
            _safe_ratio(
                pl.col("long_liq_notional") - pl.col("short_liq_notional"),
                pl.col("total_liq_notional"),
            ).alias("liq_asymmetry"),
            _safe_ratio(pl.col("oi_value"), pl.col("depth_total_notional")).alias(
                "oi_concentration"
            ),
            _safe_ratio(
                (pl.col("buy_volume") - pl.col("sell_volume")).abs(),
                pl.col("volume"),
            ).alias("vpin_bucket"),
            _safe_ratio(
                pl.col("large_trade_notional"),
                pl.col("trade_notional"),
            ).alias("large_trade_ratio"),
            (pl.col("oi_value").pct_change(60) * 100.0).alias("oi_change_1h"),
            (pl.col("oi_value").pct_change(240) * 100.0).alias("oi_change_4h"),
            (pl.col("oi_value").pct_change(1_440) * 100.0).alias("oi_change_24h"),
            (((pl.col("mid_price") / pl.col("mid_price").shift(60)) - 1.0) * 10_000.0).alias(
                "price_change_1h"
            ),
            pl.col("oi_value").diff(60).diff().alias("oi_acceleration"),
            (pl.col("spread_abs") / pl.col("mid_price") * 10_000.0).alias("spread_bps"),
            pl.col("depth_imbalance").diff().alias("depth_imbalance_momentum"),
            pl.col("cum_delta_30").diff().diff().alias("cum_delta_acceleration"),
            _safe_ratio(
                pl.col("trade_notional"),
                pl.col("mid_price") * pl.col("return_1m").abs(),
            ).alias("absorption_ratio"),
        ]
    )
    base = base.with_columns(
        [
            pl.col("carry_annualized_pct").diff().alias("carry_momentum_1h"),
            (pl.col("carry_annualized_pct") - pl.col("carry_annualized_pct").shift(24 * 60)).alias(
                "carry_momentum_24h"
            ),
            pl.col("total_liq_notional")
            .rolling_mean(window_size=60, min_samples=10)
            .alias("liq_intensity_1h"),
            pl.col("vpin_bucket").rolling_mean(window_size=50, min_samples=20).alias("vpin_50"),
            _rolling_autocorr("order_flow_imbalance", 60).alias("ofi_persistence"),
            _rolling_autocorr("return_1m_sq", 60).alias("vol_clustering"),
            _safe_ratio(pl.col("rv_5s"), pl.col("rv_5m")).alias("vol_signature_ratio"),
            (
                (pl.col("oi_change_1h").sign() != pl.col("price_change_1h").sign()).cast(pl.Float64)
                * pl.col("oi_change_1h").abs()
            ).alias("oi_price_divergence"),
            _lead_lag_score().alias("lead_lag_score"),
            (
                (pl.col("venue_dislocation_bps").abs() * 60.0)
                / (
                    pl.col("venue_dislocation_bps")
                    .diff()
                    .abs()
                    .rolling_mean(window_size=60, min_samples=10)
                    + 1e-6
                )
            )
            .clip(60.0, 3_600.0)
            .alias("convergence_half_life_secs"),
            (
                pl.when(pl.col("depth_delta") > 0.0)
                .then(pl.col("depth_delta"))
                .otherwise(0.0)
                .rolling_mean(window_size=30, min_samples=10)
                / (
                    pl.col("aggressive_flow_notional")
                    .shift(1)
                    .rolling_mean(window_size=30, min_samples=10)
                    + 1e-6
                )
            ).alias("depth_recovery_speed"),
            (
                pl.when(pl.col("cum_delta_acceleration").sign() != pl.col("price_change_1h").sign())
                .then(pl.col("cum_delta_acceleration").abs())
                .otherwise(0.0)
            ).alias("delta_divergence"),
        ]
    )
    base = base.with_columns(
        [
            _rolling_zscore("carry_annualized_pct", 7 * 24 * 60, min_samples=180).alias(
                "carry_zscore_7d"
            ),
            _rolling_zscore("carry_annualized_pct", 30 * 24 * 60, min_samples=360).alias(
                "carry_zscore_30d"
            ),
            _rolling_zscore("total_liq_notional", 24 * 60, min_samples=120).alias(
                "liq_intensity_zscore"
            ),
            _rolling_zscore("vpin_50", 4 * 60, min_samples=40).alias("toxic_flow_zscore_4h"),
            _rolling_zscore("vpin_50", 24 * 60, min_samples=120).alias("toxic_flow_zscore_24h"),
            _rolling_zscore("spread_bps", 4 * 60, min_samples=40).alias("spread_zscore_4h"),
            _rolling_zscore("spread_bps", 24 * 60, min_samples=120).alias("spread_zscore_24h"),
            _rolling_zscore("rv_1m", 24 * 60, min_samples=120).alias("rv_zscore"),
        ]
    )
    base = base.with_columns(
        [
            (
                0.5 * pl.col("oi_concentration").clip(0.0, 5.0)
                + 0.3 * pl.col("liq_intensity_zscore").fill_null(0.0).clip(-3.0, 3.0)
                + 0.2 * pl.col("carry_zscore_7d").fill_null(0.0).abs().clip(0.0, 3.0)
            ).alias("crowding_score"),
        ]
    )
    base = base.with_columns(
        [
            (
                0.5 * pl.col("crowding_score").fill_null(0.0)
                + 0.3 * pl.col("liq_intensity_zscore").fill_null(0.0).clip(-3.0, 3.0)
                + 0.2 * pl.col("liq_asymmetry").fill_null(0.0).abs()
            ).alias("cascade_risk_score"),
        ]
    )
    base = base.with_columns(
        [
            _regime_expr(
                "cascade_risk_score",
                (1.0, 2.0, 3.0),
                ("low", "elevated", "high", "extreme"),
            ).alias("cascade_regime"),
            _regime_expr(
                "toxic_flow_zscore_4h",
                (0.5, 1.5, 2.5),
                ("calm", "elevated", "toxic", "extreme"),
            ).alias("toxicity_regime"),
            pl.when(pl.col("oi_change_1h").abs() < 1.0)
            .then(pl.lit("steady"))
            .when((pl.col("oi_change_1h") > 0) & (pl.col("price_change_1h").abs() < 25.0))
            .then(pl.lit("building"))
            .otherwise(pl.lit("unwinding"))
            .alias("leverage_regime"),
            pl.when(pl.col("lead_lag_score") > 0.0001)
            .then(pl.lit("binance"))
            .when(pl.col("lead_lag_score") < -0.0001)
            .then(pl.lit("bybit"))
            .otherwise(pl.lit("neutral"))
            .alias("price_discovery_leader"),
            _regime_expr(
                "spread_zscore_4h",
                (0.5, 1.5, 2.5),
                ("tight", "normal", "wide", "stress"),
            ).alias("spread_regime"),
            _regime_expr(
                "rv_zscore",
                (0.0, 1.0, 2.0),
                ("low", "normal", "elevated", "extreme"),
            ).alias("rv_regime"),
            pl.when((pl.col("cum_delta_acceleration") > 0) & (pl.col("absorption_ratio") < 1_000.0))
            .then(pl.lit("aggressive_buying"))
            .when((pl.col("cum_delta_acceleration") > 0) & (pl.col("absorption_ratio") >= 1_000.0))
            .then(pl.lit("passive_buying"))
            .when((pl.col("cum_delta_acceleration") < 0) & (pl.col("absorption_ratio") < 1_000.0))
            .then(pl.lit("aggressive_selling"))
            .when((pl.col("cum_delta_acceleration") < 0) & (pl.col("absorption_ratio") >= 1_000.0))
            .then(pl.lit("passive_selling"))
            .otherwise(pl.lit("neutral"))
            .alias("delta_regime"),
        ]
    )
    base = base.with_columns(
        [
            pl.when(pl.col("mid_price").is_null())
            .then(pl.lit("insufficient_history"))
            .when(
                pl.any_horizontal(
                    pl.col("basis_bps").is_null(),
                    pl.col("cross_venue_funding_spread").is_null(),
                    pl.col("lead_lag_score").is_null(),
                )
            )
            .then(pl.lit("partial_context"))
            .otherwise(pl.lit("ok"))
            .alias("quality_flag")
        ]
    )

    return base.filter((pl.col("ts") >= pl.lit(start)) & (pl.col("ts") <= pl.lit(end))).select(
        [
            "ts",
            "date",
            "asset",
            "venue_scope",
            "source",
            "basis_bps",
            "funding_rate_1h",
            "carry_annualized_pct",
            "carry_zscore_7d",
            "carry_zscore_30d",
            "carry_momentum_1h",
            "carry_momentum_24h",
            "cross_venue_funding_spread",
            "long_liq_notional",
            "short_liq_notional",
            "liq_intensity_zscore",
            "liq_asymmetry",
            "oi_concentration",
            "crowding_score",
            "cascade_risk_score",
            "cascade_regime",
            "vpin_50",
            "toxic_flow_zscore_4h",
            "toxic_flow_zscore_24h",
            "ofi_persistence",
            "large_trade_ratio",
            "toxicity_regime",
            "oi_change_1h",
            "oi_change_4h",
            "oi_change_24h",
            "price_change_1h",
            "oi_price_divergence",
            "oi_acceleration",
            "leverage_regime",
            "lead_lag_score",
            "venue_dislocation_bps",
            "convergence_half_life_secs",
            "price_discovery_leader",
            "depth_recovery_speed",
            "spread_bps",
            "spread_zscore_4h",
            "spread_zscore_24h",
            "depth_imbalance_momentum",
            "spread_regime",
            "rv_5s",
            "rv_1m",
            "rv_5m",
            "vol_signature_ratio",
            "vol_clustering",
            "rv_regime",
            "cum_delta_acceleration",
            "delta_divergence",
            "absorption_ratio",
            "delta_regime",
            "quality_flag",
        ]
    )


def build_perps_composite_series(
    store: DatasetStore,
    start: datetime,
    end: datetime,
    *,
    asset: str,
    venue_scope: str = "all",
    source: str = COMPOSITE_SOURCE,
) -> pl.DataFrame:
    signal_frame = store.read(
        DatasetName.PERPS_SIGNAL_SERIES,
        start=start,
        end=end,
        filters={"asset": asset, "venue_scope": venue_scope, "source": SOURCE},
    ).sort("ts")
    if signal_frame.is_empty():
        return pl.DataFrame()
    return signal_frame.select(
        [
            "ts",
            "date",
            pl.lit(asset).alias("asset"),
            pl.lit(venue_scope).alias("venue_scope"),
            pl.lit(source).alias("source"),
            _avg_score("carry_zscore_7d", "carry_zscore_30d").alias("carry_score"),
            _avg_score(
                "cascade_risk_score",
                "liq_intensity_zscore",
                "crowding_score",
            ).alias("cascade_risk_score"),
            _avg_score(
                "toxic_flow_zscore_4h",
                "toxic_flow_zscore_24h",
                pl.col("large_trade_ratio") * 3.0,
            ).alias("toxicity_score"),
            _avg_score(
                "oi_price_divergence",
                "oi_acceleration",
                "crowding_score",
            ).alias("positioning_score"),
            _avg_score(
                "spread_zscore_4h",
                "spread_zscore_24h",
                pl.col("venue_dislocation_bps") / 5.0,
            ).alias("fragility_score"),
            "quality_flag",
        ]
    )


def build_perps_transition_series(
    store: DatasetStore,
    start: datetime,
    end: datetime,
    *,
    asset: str,
    venue_scope: str = "all",
    source: str = TRANSITION_SOURCE,
) -> pl.DataFrame:
    frame = store.read(
        DatasetName.PERPS_SIGNAL_SERIES,
        start=start,
        end=end,
        filters={"asset": asset, "venue_scope": venue_scope, "source": SOURCE},
    ).sort("ts")
    if frame.is_empty():
        return pl.DataFrame()

    rows: list[dict[str, Any]] = []
    for signal_family, column in REGIME_FAMILIES:
        states = [str(value) for value in frame.get_column(column).to_list() if value is not None]
        if not states:
            continue
        history_transitions: Counter[tuple[str, str]] = Counter()
        history_origin: Counter[str] = Counter()
        run_lengths: list[int] = []
        last_state: str | None = None
        run_length = 0
        for row in frame.iter_rows(named=True):
            current = row.get(column)
            if current is None:
                continue
            current = str(current)
            if last_state is not None:
                history_transitions[(last_state, current)] += 1
                history_origin[last_state] += 1
            if current == last_state:
                run_length += 1
            else:
                if run_length:
                    run_lengths.append(run_length)
                run_length = 1
            distribution = {
                target: history_transitions[(current, target)] / history_origin[current]
                for target in sorted(
                    {right for left, right in history_transitions if left == current}
                )
                if history_origin[current] > 0
            }
            median_duration = (
                float(sorted(run_lengths)[len(run_lengths) // 2])
                if run_lengths
                else float(run_length)
            )
            rows.append(
                {
                    "ts": row["ts"],
                    "date": row["date"],
                    "asset": asset,
                    "venue_scope": venue_scope,
                    "source": source,
                    "signal_family": signal_family,
                    "current_state": current,
                    "prior_state": last_state,
                    "next_state_distribution": orjson.dumps(distribution).decode(),
                    "median_state_duration_minutes": median_duration,
                    "quality_flag": row.get("quality_flag") or "ok",
                }
            )
            last_state = current
        if run_length:
            run_lengths.append(run_length)
    return pl.DataFrame(rows) if rows else pl.DataFrame()


def _with_bar_context(frame: pl.DataFrame) -> pl.DataFrame:
    if frame.is_empty():
        return frame
    return frame.with_columns(
        [
            pl.coalesce([pl.col("venue_group"), _venue_group_expr("venue")]).alias("venue_group"),
            pl.coalesce([pl.col("market_type"), _market_type_expr("venue")]).alias("market_type"),
            pl.coalesce(
                [pl.col("is_derivatives"), _market_type_expr("venue") != pl.lit("spot")]
            ).alias("is_derivatives"),
            pl.coalesce([pl.col("contract_type"), _contract_type_expr("venue")]).alias(
                "contract_type"
            ),
            pl.coalesce([pl.col("exchange_symbol"), pl.col("venue")]).alias("exchange_symbol"),
            pl.coalesce(
                [
                    pl.col("depth_notional_bid_5"),
                    pl.col("depth_bid_5") * pl.col("close"),
                ]
            ).alias("depth_notional_bid_5"),
            pl.coalesce(
                [
                    pl.col("depth_notional_ask_5"),
                    pl.col("depth_ask_5") * pl.col("close"),
                ]
            ).alias("depth_notional_ask_5"),
        ]
    )


def _with_book_context(frame: pl.DataFrame) -> pl.DataFrame:
    if frame.is_empty():
        return frame
    return frame.with_columns(
        [
            pl.coalesce([pl.col("venue_group"), _venue_group_expr("venue")]).alias("venue_group"),
            pl.coalesce([pl.col("market_type"), _market_type_expr("venue")]).alias("market_type"),
            pl.coalesce(
                [pl.col("is_derivatives"), _market_type_expr("venue") != pl.lit("spot")]
            ).alias("is_derivatives"),
            pl.coalesce([pl.col("contract_type"), _contract_type_expr("venue")]).alias(
                "contract_type"
            ),
            pl.coalesce([pl.col("exchange_symbol"), pl.col("venue")]).alias("exchange_symbol"),
            pl.coalesce(
                [
                    pl.col("depth_notional_bid_5"),
                    pl.col("depth_bid_5") * pl.col("mid_price"),
                ]
            ).alias("depth_notional_bid_5"),
            pl.coalesce(
                [
                    pl.col("depth_notional_ask_5"),
                    pl.col("depth_ask_5") * pl.col("mid_price"),
                ]
            ).alias("depth_notional_ask_5"),
        ]
    )


def _with_raw_context(frame: pl.DataFrame) -> pl.DataFrame:
    if frame.is_empty():
        return frame
    return frame.with_columns(
        [
            pl.coalesce([pl.col("venue_group"), _venue_group_expr("venue")]).alias("venue_group"),
            pl.coalesce([pl.col("market_type"), _market_type_expr("venue")]).alias("market_type"),
            pl.coalesce(
                [pl.col("is_derivatives"), _market_type_expr("venue") != pl.lit("spot")]
            ).alias("is_derivatives"),
            pl.coalesce([pl.col("contract_type"), _contract_type_expr("venue")]).alias(
                "contract_type"
            ),
            pl.coalesce([pl.col("exchange_symbol"), pl.col("venue")]).alias("exchange_symbol"),
            pl.coalesce([pl.col("liquidation_side"), pl.col("side")]).alias("liquidation_side"),
            pl.coalesce([pl.col("trade_notional"), pl.col("price") * pl.col("size")]).alias(
                "trade_notional"
            ),
            pl.coalesce(
                [
                    pl.col("depth_notional_bid_5"),
                    pl.col("bid") * pl.col("size"),
                ]
            ).alias("depth_notional_bid_5"),
            pl.coalesce(
                [
                    pl.col("depth_notional_ask_5"),
                    pl.col("ask") * pl.col("size"),
                ]
            ).alias("depth_notional_ask_5"),
        ]
    )


def _with_funding_context(frame: pl.DataFrame) -> pl.DataFrame:
    if frame.is_empty():
        return frame
    return frame.with_columns([_venue_group_expr("venue").alias("venue_group")])


def _read_hot_window(
    store: DatasetStore,
    dataset: DatasetName,
    start: datetime,
    end: datetime,
    *,
    filters: dict[str, Any],
) -> pl.DataFrame:
    if start.date() == end.date():
        requested_columns = _HOT_COLUMNS.get(dataset)
        if requested_columns:
            return _read_hot_columns(
                store,
                dataset,
                start=start,
                end=end,
                filters=filters,
                columns=requested_columns,
            )
        else:
            return store.read(
                dataset,
                start=start,
                end=end,
                filters=filters,
                read_mode=ReadMode.LATEST,
            )
    return store.read(dataset, start=start, end=end, filters=filters)


def _read_hot_columns(
    store: DatasetStore,
    dataset: DatasetName,
    *,
    start: datetime,
    end: datetime,
    filters: dict[str, Any],
    columns: tuple[str, ...],
) -> pl.DataFrame:
    spec = DATASET_SPECS[dataset]
    required = set(columns) | set(spec.primary_key) | {"run_ts"}
    ordered = [column for column in spec.schema if column in required]
    base_frame = store.read(
        dataset,
        start=start,
        end=end,
        filters=filters,
        read_mode=ReadMode.LATEST,
    )
    available_columns = set(base_frame.columns)
    frame = base_frame.select([pl.col(column) for column in ordered if column in available_columns])
    missing = [column for column in ordered if column not in frame.columns]
    if missing:
        frame = frame.with_columns(
            [pl.lit(None, dtype=spec.schema[column]).alias(column) for column in missing]
        )
    if frame.is_empty():
        return empty_frame(dataset)
    if filters:
        for column, value in filters.items():
            if column in frame.columns:
                frame = frame.filter(pl.col(column) == value)
    return frame


_HOT_COLUMNS: dict[DatasetName, tuple[str, ...]] = {
    DatasetName.MICROSTRUCTURE_BAR: (
        "ts",
        "date",
        "asset",
        "venue",
        "exchange_symbol",
        "venue_group",
        "market_type",
        "is_derivatives",
        "contract_type",
        "clock",
        "bar_size_secs",
        "close",
        "volume",
        "buy_volume",
        "sell_volume",
        "trade_count",
        "spread",
        "depth_bid_5",
        "depth_ask_5",
        "depth_notional_bid_5",
        "depth_notional_ask_5",
        "run_ts",
    ),
    DatasetName.MICROSTRUCTURE_BOOK_STATE: (
        "ts",
        "date",
        "asset",
        "venue",
        "exchange_symbol",
        "venue_group",
        "market_type",
        "is_derivatives",
        "contract_type",
        "mid_price",
        "depth_bid_5",
        "depth_ask_5",
        "depth_notional_bid_5",
        "depth_notional_ask_5",
        "run_ts",
    ),
    DatasetName.MICROSTRUCTURE_RAW_EVENT: (
        "ts",
        "date",
        "asset",
        "venue",
        "event_type",
        "exchange_symbol",
        "venue_group",
        "market_type",
        "is_derivatives",
        "contract_type",
        "side",
        "liquidation_side",
        "price",
        "size",
        "trade_notional",
        "bid",
        "ask",
        "mark_price",
        "open_interest",
        "depth_notional_bid_5",
        "depth_notional_ask_5",
        "run_ts",
    ),
    DatasetName.FUNDING_RATE_SNAPSHOT: (
        "ts",
        "date",
        "venue",
        "underlying",
        "interest_1h",
        "run_ts",
    ),
}


def _filter_scope(frame: pl.DataFrame, venue_scope: str) -> pl.DataFrame:
    if frame.is_empty() or venue_scope == "all":
        return frame
    if venue_scope == "binance":
        return frame.filter(pl.col("venue_group") == "binance")
    if venue_scope == "bybit":
        return frame.filter(pl.col("venue_group") == "bybit")
    return frame


def _filter_funding_scope(frame: pl.DataFrame, venue_scope: str) -> pl.DataFrame:
    if frame.is_empty() or venue_scope == "all":
        return frame
    return frame.filter(pl.col("venue_group") == venue_scope)


def _aggregate_asset_bars(frame: pl.DataFrame) -> pl.DataFrame:
    if frame.is_empty():
        return pl.DataFrame()
    grouped = (
        frame.sort(["ts", "venue"])
        .group_by("ts")
        .agg(
            [
                pl.col("date").first().alias("date"),
                pl.col("close").mean().alias("mid_price"),
                pl.col("close").filter(~pl.col("is_derivatives")).mean().alias("spot_mid"),
                pl.col("close").filter(pl.col("is_derivatives")).mean().alias("perp_mid"),
                pl.col("close")
                .filter((pl.col("venue_group") == "binance") & pl.col("is_derivatives"))
                .mean()
                .alias("binance_mid"),
                pl.col("close")
                .filter((pl.col("venue_group") == "bybit") & pl.col("is_derivatives"))
                .mean()
                .alias("bybit_mid"),
                pl.col("volume").sum().alias("volume"),
                pl.col("buy_volume").sum().alias("buy_volume"),
                pl.col("sell_volume").sum().alias("sell_volume"),
                pl.col("trade_count").sum().alias("trade_count"),
                pl.col("spread").mean().alias("spread_abs"),
                pl.col("depth_notional_bid_5").sum().alias("depth_notional_bid_5"),
                pl.col("depth_notional_ask_5").sum().alias("depth_notional_ask_5"),
            ]
        )
    )
    return grouped.with_columns(
        [
            (pl.col("depth_notional_bid_5") + pl.col("depth_notional_ask_5")).alias(
                "depth_total_notional"
            ),
            _safe_ratio(
                pl.col("depth_notional_bid_5") - pl.col("depth_notional_ask_5"),
                pl.col("depth_notional_bid_5") + pl.col("depth_notional_ask_5"),
            ).alias("depth_imbalance"),
            (
                (pl.col("buy_volume") - pl.col("sell_volume")).rolling_sum(
                    window_size=10, min_samples=3
                )
            ).alias("cum_delta_10"),
            (
                (pl.col("buy_volume") - pl.col("sell_volume")).rolling_sum(
                    window_size=30, min_samples=5
                )
            ).alias("cum_delta_30"),
            ((pl.col("mid_price") / pl.col("mid_price").shift(1)).log()).alias("return_1m"),
            ((pl.col("mid_price") / pl.col("mid_price").shift(1)).log() ** 2).alias("return_1m_sq"),
            ((pl.col("buy_volume") - pl.col("sell_volume")) / (pl.col("volume") + 1e-6)).alias(
                "order_flow_imbalance"
            ),
        ]
    )


def _aggregate_trade_context(frame: pl.DataFrame) -> pl.DataFrame:
    trades = frame.filter(pl.col("event_type") == "trade").sort("ts")
    if trades.is_empty():
        return pl.DataFrame()
    trade_notional = trades.get_column("trade_notional")
    mean_value = _numeric_or_zero(trade_notional.mean())
    std_value = _numeric_or_zero(trade_notional.std())
    threshold = mean_value + 2.0 * std_value
    return (
        trades.group_by_dynamic("ts", every="1m", closed="left")
        .agg(
            [
                pl.col("trade_notional").sum().alias("trade_notional"),
                pl.when(pl.col("trade_notional") >= threshold)
                .then(pl.col("trade_notional"))
                .otherwise(0.0)
                .sum()
                .alias("large_trade_notional"),
                pl.when(pl.col("side") == "buy")
                .then(pl.col("trade_notional"))
                .otherwise(0.0)
                .sum()
                .alias("aggressive_buy_notional"),
                pl.when(pl.col("side") == "sell")
                .then(pl.col("trade_notional"))
                .otherwise(0.0)
                .sum()
                .alias("aggressive_sell_notional"),
            ]
        )
        .with_columns(
            [
                (pl.col("aggressive_buy_notional") + pl.col("aggressive_sell_notional")).alias(
                    "aggressive_flow_notional"
                )
            ]
        )
    )


def _aggregate_liquidation_context(frame: pl.DataFrame) -> pl.DataFrame:
    liquidations = frame.filter(pl.col("event_type") == "liquidation").sort("ts")
    if liquidations.is_empty():
        return pl.DataFrame()
    return liquidations.group_by_dynamic("ts", every="1m", closed="left").agg(
        [
            pl.when(pl.col("liquidation_side") == "long")
            .then(pl.col("trade_notional"))
            .otherwise(0.0)
            .sum()
            .alias("long_liq_notional"),
            pl.when(pl.col("liquidation_side") == "short")
            .then(pl.col("trade_notional"))
            .otherwise(0.0)
            .sum()
            .alias("short_liq_notional"),
            pl.col("trade_notional").sum().alias("total_liq_notional"),
        ]
    )


def _aggregate_open_interest_context(frame: pl.DataFrame) -> pl.DataFrame:
    oi = frame.filter(pl.col("open_interest").is_not_null() & pl.col("is_derivatives")).sort("ts")
    if oi.is_empty():
        return pl.DataFrame()
    return oi.group_by_dynamic("ts", every="1m", closed="left").agg(
        [pl.col("open_interest").last().alias("oi_value")]
    )


def _aggregate_funding_context(frame: pl.DataFrame) -> pl.DataFrame:
    if frame.is_empty():
        return pl.DataFrame()
    grouped = (
        frame.sort("ts")
        .group_by_dynamic(
            "ts",
            every="1m",
            group_by=["venue_group"],
            closed="left",
        )
        .agg([pl.col("interest_1h").last().alias("funding_rate_1h")])
    )
    asset = grouped.group_by("ts").agg([pl.col("funding_rate_1h").mean().alias("funding_rate_1h")])
    by_group = grouped.pivot(
        index="ts",
        on="venue_group",
        values="funding_rate_1h",
        aggregate_function="last",
    )
    if not by_group.is_empty():
        asset = asset.join(by_group, on="ts", how="left")
    if "binance" not in asset.columns:
        asset = asset.with_columns(pl.lit(None, dtype=pl.Float64).alias("binance"))
    if "bybit" not in asset.columns:
        asset = asset.with_columns(pl.lit(None, dtype=pl.Float64).alias("bybit"))
    return asset.with_columns(
        [(pl.col("binance") - pl.col("bybit")).alias("cross_venue_funding_spread")]
    ).drop(["binance", "bybit"])


def _aggregate_cross_venue_context(
    frame: pl.DataFrame, _funding_frame: pl.DataFrame
) -> pl.DataFrame:
    if frame.is_empty():
        return pl.DataFrame()
    grouped = (
        frame.group_by("ts")
        .agg(
            [
                pl.col("close")
                .filter((pl.col("venue_group") == "binance") & pl.col("is_derivatives"))
                .mean()
                .alias("binance_mid"),
                pl.col("close")
                .filter((pl.col("venue_group") == "bybit") & pl.col("is_derivatives"))
                .mean()
                .alias("bybit_mid"),
            ]
        )
        .sort("ts")
        .with_columns(
            [
                (
                    ((pl.col("binance_mid") - pl.col("bybit_mid")).abs())
                    / ((pl.col("binance_mid") + pl.col("bybit_mid")) / 2.0)
                    * 10_000.0
                ).alias("venue_dislocation_bps"),
                ((pl.col("binance_mid") / pl.col("binance_mid").shift(1)).log()).alias(
                    "binance_ret"
                ),
                ((pl.col("bybit_mid") / pl.col("bybit_mid").shift(1)).log()).alias("bybit_ret"),
            ]
        )
    )
    return grouped


def _aggregate_volatility_context(bars_5s: pl.DataFrame, bars_1m: pl.DataFrame) -> pl.DataFrame:
    rows: list[pl.DataFrame] = []
    if not bars_5s.is_empty():
        frame_5s = (
            bars_5s.group_by("ts")
            .agg(pl.col("close").mean().alias("mid_price"))
            .sort("ts")
            .with_columns(
                ((pl.col("mid_price") / pl.col("mid_price").shift(1)).log()).alias("ret_5s")
            )
            .with_columns(
                [
                    (
                        pl.col("ret_5s").rolling_std(window_size=720, min_samples=120)
                        * math.sqrt(SECONDS_PER_YEAR / 5.0)
                    ).alias("rv_5s"),
                ]
            )
        )
        rows.append(
            frame_5s.group_by_dynamic("ts", every="1m", closed="left").agg(
                pl.col("rv_5s").last().alias("rv_5s")
            )
        )
    if not bars_1m.is_empty():
        frame_1m = (
            bars_1m.group_by("ts")
            .agg(pl.col("close").mean().alias("mid_price"))
            .sort("ts")
            .with_columns(
                ((pl.col("mid_price") / pl.col("mid_price").shift(1)).log()).alias("ret_1m")
            )
            .with_columns(
                [
                    (
                        pl.col("ret_1m").rolling_std(window_size=60, min_samples=15)
                        * math.sqrt(SECONDS_PER_YEAR / 60.0)
                    ).alias("rv_1m"),
                ]
            )
        )
        frame_5m = (
            frame_1m.group_by_dynamic("ts", every="5m", closed="left")
            .agg(pl.col("mid_price").last().alias("mid_price_5m"))
            .sort("ts")
            .with_columns(
                ((pl.col("mid_price_5m") / pl.col("mid_price_5m").shift(1)).log()).alias("ret_5m")
            )
            .with_columns(
                [
                    (
                        pl.col("ret_5m").rolling_std(window_size=12, min_samples=6)
                        * math.sqrt(SECONDS_PER_YEAR / 300.0)
                    ).alias("rv_5m"),
                ]
            )
        )
        rows.append(frame_1m.select(["ts", "rv_1m"]))
        rows.append(frame_5m.select(["ts", "rv_5m"]))
    if not rows:
        return pl.DataFrame()
    out = rows[0]
    for frame in rows[1:]:
        out = out.join(frame, on="ts", how="full", coalesce=True)
    return out.sort("ts")


def _aggregate_depth_context(book: pl.DataFrame, raw_events: pl.DataFrame) -> pl.DataFrame:
    if book.is_empty():
        return pl.DataFrame()
    grouped = (
        book.group_by_dynamic("ts", every="1m", closed="left")
        .agg(
            [
                (pl.col("depth_notional_bid_5") + pl.col("depth_notional_ask_5"))
                .mean()
                .alias("depth_total_notional"),
                _safe_ratio(
                    pl.col("depth_notional_bid_5") - pl.col("depth_notional_ask_5"),
                    pl.col("depth_notional_bid_5") + pl.col("depth_notional_ask_5"),
                )
                .mean()
                .alias("depth_imbalance"),
            ]
        )
        .sort("ts")
        .with_columns(pl.col("depth_total_notional").diff().alias("depth_delta"))
    )
    aggressive = _aggregate_trade_context(raw_events)
    if aggressive.is_empty():
        return grouped
    return grouped.join(aggressive.select(["ts", "aggressive_flow_notional"]), on="ts", how="left")


def _venue_group_expr(column: str) -> pl.Expr:
    return (
        pl.when(pl.col(column).str.contains("binance"))
        .then(pl.lit("binance"))
        .when(pl.col(column).str.contains("bybit"))
        .then(pl.lit("bybit"))
        .when(pl.col(column).str.contains("deribit"))
        .then(pl.lit("deribit"))
        .otherwise(pl.lit("other"))
    )


def _market_type_expr(column: str) -> pl.Expr:
    return (
        pl.when(pl.col(column).str.contains("spot"))
        .then(pl.lit("spot"))
        .when(pl.col(column).str.contains("futures"))
        .then(pl.lit("futures"))
        .when(pl.col(column).str.contains("perps"))
        .then(pl.lit("perp"))
        .otherwise(pl.lit("perp"))
    )


def _contract_type_expr(column: str) -> pl.Expr:
    return (
        pl.when(pl.col(column).str.contains("spot"))
        .then(pl.lit("spot"))
        .when(pl.col(column).str.contains("futures"))
        .then(pl.lit("linear_future"))
        .when(pl.col(column).str.contains("perps"))
        .then(pl.lit("linear_perp"))
        .otherwise(pl.lit("linear_perp"))
    )


def _rolling_zscore(column: str, window: int, *, min_samples: int) -> pl.Expr:
    mean = pl.col(column).rolling_mean(window_size=window, min_samples=min_samples)
    std = pl.col(column).rolling_std(window_size=window, min_samples=min_samples)
    return (pl.col(column) - mean) / (std + 1e-9)


def _rolling_autocorr(column: str, window: int) -> pl.Expr:
    numerator = (pl.col(column) * pl.col(column).shift(1)).rolling_mean(
        window_size=window, min_samples=max(5, window // 4)
    )
    denominator = (pl.col(column) ** 2).rolling_mean(
        window_size=window,
        min_samples=max(5, window // 4),
    )
    return numerator / (denominator + 1e-9)


def _safe_ratio(numerator: pl.Expr, denominator: pl.Expr) -> pl.Expr:
    return numerator / (denominator + 1e-9)


def _lead_lag_score() -> pl.Expr:
    lead_binance = (pl.col("binance_ret").shift(1) * pl.col("bybit_ret")).rolling_mean(
        window_size=120, min_samples=20
    )
    lead_bybit = (pl.col("bybit_ret").shift(1) * pl.col("binance_ret")).rolling_mean(
        window_size=120, min_samples=20
    )
    return lead_binance - lead_bybit


def _regime_expr(
    column: str, thresholds: tuple[float, float, float], labels: tuple[str, str, str, str]
) -> pl.Expr:
    low, medium, high = thresholds
    l0, l1, l2, l3 = labels
    return (
        pl.when(pl.col(column).is_null())
        .then(pl.lit(l0))
        .when(pl.col(column) >= high)
        .then(pl.lit(l3))
        .when(pl.col(column) >= medium)
        .then(pl.lit(l2))
        .when(pl.col(column) >= low)
        .then(pl.lit(l1))
        .otherwise(pl.lit(l0))
    )


def _avg_score(*columns: pl.Expr | str) -> pl.Expr:
    exprs: list[pl.Expr] = []
    for column in columns:
        if isinstance(column, str):
            exprs.append(pl.col(column))
        else:
            exprs.append(column)
    total = pl.sum_horizontal([expr.fill_null(0.0).clip(-5.0, 5.0) for expr in exprs])
    count = pl.sum_horizontal(
        [pl.when(expr.is_not_null()).then(1.0).otherwise(0.0) for expr in exprs]
    )
    return total / (count + 1e-9)


def _numeric_or_zero(value: Any) -> float:
    if isinstance(value, bool):
        return float(value)
    if isinstance(value, (int, float)):
        return float(value)
    return 0.0



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import Any

import numpy as np
import polars as pl
from sklearn.isotonic import IsotonicRegression
from sklearn.linear_model import Ridge

from uforia.storage import DATASET_SPECS, DatasetName, DatasetStore
from uforia.storage.datasets import empty_frame

SOURCE = "composite_v1"
MARKET_ASSET = "MARKET"
MARKET_SCOPE = "market"
ASSET_SCOPE = "asset"
LOOKBACK_WINDOW_HOURS = 45 * 24
NORMALIZATION_WINDOW_HOURS = 14 * 24
NORMALIZATION_MIN_SAMPLES = 24
NORMALIZATION_FALLBACK_MIN_SAMPLES = 6
CALIBRATION_MIN_SAMPLES = 48
CALIBRATION_WINDOW_HOURS = 30 * 24
MARKET_WEIGHTS = {"BTC": 0.65, "ETH": 0.35}
HORIZON_WEIGHTS = {"30s": 0.15, "1m": 0.30, "2m": 0.25, "5m": 0.30}
PILLAR_WEIGHTS = {
    "sentiment": {"options": 0.40, "perps": 0.40, "micro": 0.20},
    "directional": {"micro": 0.60, "perps": 0.30, "options": 0.10},
    "fragility": {"perps": 0.40, "options": 0.40, "micro": 0.20},
}


@dataclass(frozen=True)
class ComponentSpec:
    pillar: str
    family: str
    component: str
    source_column: str
    sign: float = 1.0
    weight: float = 1.0


COMPONENT_SPECS: tuple[ComponentSpec, ...] = (
    ComponentSpec("sentiment", "options", "rvvix_14d", "vvix_rvvix_14d", sign=-1.0),
    ComponentSpec("sentiment", "options", "qvvix_30d", "vvix_qvvix_30d", sign=-1.0),
    ComponentSpec("sentiment", "options", "razor_pnl_zscore", "razor_razor_pnl_zscore"),
    ComponentSpec("sentiment", "options", "razor_vrp_zscore", "razor_razor_vrp_zscore"),
    ComponentSpec("sentiment", "options", "vol_carry_score", "options_vol_carry_score"),
    ComponentSpec(
        "sentiment",
        "options",
        "skew_stress_score",
        "options_skew_stress_score",
        sign=-1.0,
    ),
    ComponentSpec(
        "sentiment",
        "options",
        "ml_carry_toxicity_score",
        "options_ml_carry_toxicity_score",
    ),
    ComponentSpec(
        "sentiment", "options", "funding_skew_score", "options_funding_skew_score"
    ),
    ComponentSpec("sentiment", "options", "sessionality_score", "options_sessionality_score"),
    ComponentSpec("sentiment", "options", "dispersion_score", "options_dispersion_score"),
    ComponentSpec("sentiment", "options", "expiry_pin_score", "options_expiry_pin_score"),
    ComponentSpec(
        "sentiment", "options", "gamma_flip_score", "options_gamma_flip_score", sign=-1.0
    ),
    ComponentSpec("sentiment", "perps", "carry_score", "perps_composite_carry_score"),
    ComponentSpec(
        "sentiment", "perps", "positioning_score", "perps_composite_positioning_score"
    ),
    ComponentSpec(
        "sentiment",
        "perps",
        "cascade_risk_score",
        "perps_composite_cascade_risk_score",
        sign=-1.0,
    ),
    ComponentSpec(
        "sentiment", "perps", "fragility_score", "perps_composite_fragility_score", sign=-1.0
    ),
    ComponentSpec(
        "sentiment", "perps", "toxicity_score", "perps_composite_toxicity_score", sign=-1.0
    ),
    ComponentSpec("sentiment", "micro", "micro_sentiment", "micro_sentiment_raw"),
    ComponentSpec("directional", "micro", "micro_direction", "micro_direction_raw"),
    ComponentSpec(
        "directional", "perps", "carry_momentum_1h", "perps_signal_carry_momentum_1h"
    ),
    ComponentSpec(
        "directional", "perps", "carry_momentum_24h", "perps_signal_carry_momentum_24h"
    ),
    ComponentSpec(
        "directional",
        "perps",
        "cum_delta_acceleration",
        "perps_signal_cum_delta_acceleration",
    ),
    ComponentSpec(
        "directional",
        "perps",
        "depth_imbalance_momentum",
        "perps_signal_depth_imbalance_momentum",
    ),
    ComponentSpec("directional", "perps", "lead_lag_score", "perps_signal_lead_lag_score"),
    ComponentSpec(
        "directional", "perps", "depth_recovery_speed", "perps_signal_depth_recovery_speed"
    ),
    ComponentSpec(
        "directional",
        "perps",
        "oi_price_divergence",
        "perps_signal_oi_price_divergence",
        sign=-1.0,
        weight=0.5,
    ),
    ComponentSpec(
        "directional",
        "perps",
        "delta_divergence",
        "perps_signal_delta_divergence",
        sign=-1.0,
        weight=0.5,
    ),
    ComponentSpec(
        "directional", "options", "funding_skew_score", "options_funding_skew_score"
    ),
    ComponentSpec("directional", "options", "gamma_flip_score", "options_gamma_flip_score"),
    ComponentSpec(
        "directional", "options", "sessionality_score", "options_sessionality_score", weight=0.5
    ),
    ComponentSpec("fragility", "options", "rvvix_14d", "vvix_rvvix_14d"),
    ComponentSpec("fragility", "options", "qvvix_30d", "vvix_qvvix_30d"),
    ComponentSpec(
        "fragility", "options", "skew_stress_score", "options_skew_stress_score"
    ),
    ComponentSpec("fragility", "options", "gamma_flip_score", "options_gamma_flip_score"),
    ComponentSpec(
        "fragility",
        "perps",
        "cascade_risk_score",
        "perps_composite_cascade_risk_score",
    ),
    ComponentSpec("fragility", "perps", "toxicity_score", "perps_composite_toxicity_score"),
    ComponentSpec("fragility", "perps", "fragility_score", "perps_composite_fragility_score"),
    ComponentSpec("fragility", "perps", "spread_zscore_4h", "perps_signal_spread_zscore_4h"),
    ComponentSpec("fragility", "perps", "oi_concentration", "perps_signal_oi_concentration"),
    ComponentSpec("fragility", "micro", "micro_fragility", "micro_fragility_raw"),
)

SENTIMENT_REGIMES = ((20.0, "panic"), (40.0, "fear"), (60.0, "neutral"), (80.0, "greed"))
DIRECTIONAL_REGIMES = (
    (-40.0, "bearish"),
    (-15.0, "mild_bearish"),
    (15.0, "neutral"),
    (40.0, "mild_bullish"),
)
FRAGILITY_REGIMES = ((25.0, "stable"), (50.0, "watch"), (75.0, "fragile"))
REGIME_ORDER = {
    "sentiment": ["panic", "fear", "neutral", "greed", "euphoria"],
    "directional": ["bearish", "mild_bearish", "neutral", "mild_bullish", "bullish"],
    "fragility": ["stable", "watch", "fragile", "squeeze_risk"],
}


def _frame_from_rows(dataset: DatasetName, rows: list[dict[str, Any]]) -> pl.DataFrame:
    if not rows:
        return empty_frame(dataset)
    return pl.DataFrame(rows, schema=DATASET_SPECS[dataset].schema, strict=False)


def build_composite_signal_series(
    store: DatasetStore,
    start: datetime,
    end: datetime,
    *,
    asset: str,
    source: str = SOURCE,
) -> pl.DataFrame:
    lookback_start = start - timedelta(hours=LOOKBACK_WINDOW_HOURS)
    raw_frame = _build_asset_raw_frame(store, lookback_start, end, asset.upper())
    if raw_frame.is_empty():
        return empty_frame(DatasetName.COMPOSITE_SIGNAL_SERIES)

    normalized = _normalize_components(raw_frame)
    component_frame = _build_component_frame(
        normalized, asset=asset.upper(), scope=ASSET_SCOPE, source=source
    )
    signal_frame = _build_signal_frame(
        normalized, component_frame, asset=asset.upper(), scope=ASSET_SCOPE, source=source
    )
    calibrated = _apply_calibration(signal_frame, raw_frame)
    return calibrated.filter((pl.col("ts") >= pl.lit(start)) & (pl.col("ts") <= pl.lit(end)))


def build_market_composite_signal_series(
    store: DatasetStore,
    start: datetime,
    end: datetime,
    *,
    source: str = SOURCE,
) -> pl.DataFrame:
    assets = {
        asset: store.read(
            DatasetName.COMPOSITE_SIGNAL_SERIES,
            start=start,
            end=end,
            filters={"asset": asset, "scope": ASSET_SCOPE, "source": source},
        )
        for asset in MARKET_WEIGHTS
    }
    if any(frame.is_empty() for frame in assets.values()):
        return empty_frame(DatasetName.COMPOSITE_SIGNAL_SERIES)
    rows: list[dict[str, Any]] = []
    joined = _join_market_frames(assets)
    for item in joined.iter_rows(named=True):
        sentiment = _weighted_market_value(item, "sentiment_index")
        directional = _weighted_market_value(item, "directional_tilt")
        fragility = _weighted_market_value(item, "fragility_index")
        confidence = _weighted_market_value(item, "confidence_score")
        rows.append(
            {
                "ts": item["ts"],
                "date": item["ts"].date(),
                "scope": MARKET_SCOPE,
                "asset": MARKET_ASSET,
                "source": source,
                "sentiment_index": sentiment,
                "directional_tilt": directional,
                "fragility_index": fragility,
                "confidence_score": confidence,
                "options_sentiment_subindex": _weighted_market_value(
                    item, "options_sentiment_subindex"
                ),
                "perps_sentiment_subindex": _weighted_market_value(
                    item, "perps_sentiment_subindex"
                ),
                "micro_sentiment_subindex": _weighted_market_value(
                    item, "micro_sentiment_subindex"
                ),
                "micro_direction_subindex": _weighted_market_value(
                    item, "micro_direction_subindex"
                ),
                "perps_direction_subindex": _weighted_market_value(
                    item, "perps_direction_subindex"
                ),
                "options_direction_subindex": _weighted_market_value(
                    item, "options_direction_subindex"
                ),
                "perps_fragility_subindex": _weighted_market_value(
                    item, "perps_fragility_subindex"
                ),
                "options_fragility_subindex": _weighted_market_value(
                    item, "options_fragility_subindex"
                ),
                "micro_fragility_subindex": _weighted_market_value(
                    item, "micro_fragility_subindex"
                ),
                "coverage_ratio": _weighted_market_value(item, "coverage_ratio"),
                "freshness_score": _weighted_market_value(item, "freshness_score"),
                "calibrator_mode": "market_weighted",
                "sentiment_regime": _sentiment_regime(sentiment),
                "directional_regime": _directional_regime(directional),
                "fragility_regime": _fragility_regime(fragility),
                "quality_flag": _quality_flag(
                    _weighted_market_value(item, "coverage_ratio"),
                    _weighted_market_value(item, "confidence_score"),
                ),
            }
        )
    if not rows:
        return empty_frame(DatasetName.COMPOSITE_SIGNAL_SERIES)
    return _frame_from_rows(DatasetName.COMPOSITE_SIGNAL_SERIES, rows)


def build_composite_component_series(
    store: DatasetStore,
    start: datetime,
    end: datetime,
    *,
    asset: str,
    source: str = SOURCE,
) -> pl.DataFrame:
    lookback_start = start - timedelta(hours=LOOKBACK_WINDOW_HOURS)
    raw_frame = _build_asset_raw_frame(store, lookback_start, end, asset.upper())
    if raw_frame.is_empty():
        return empty_frame(DatasetName.COMPOSITE_COMPONENT_SERIES)
    normalized = _normalize_components(raw_frame)
    return _build_component_frame(
        normalized,
        asset=asset.upper(),
        scope=ASSET_SCOPE,
        source=source,
    ).filter((pl.col("ts") >= pl.lit(start)) & (pl.col("ts") <= pl.lit(end)))


def build_market_composite_component_series(
    store: DatasetStore,
    start: datetime,
    end: datetime,
    *,
    source: str = SOURCE,
) -> pl.DataFrame:
    frames = {
        asset: store.read(
            DatasetName.COMPOSITE_COMPONENT_SERIES,
            start=start,
            end=end,
            filters={"asset": asset, "scope": ASSET_SCOPE, "source": source},
        )
        for asset in MARKET_WEIGHTS
    }
    if any(frame.is_empty() for frame in frames.values()):
        return empty_frame(DatasetName.COMPOSITE_COMPONENT_SERIES)
    rows: list[dict[str, Any]] = []
    btc = frames["BTC"]
    eth = frames["ETH"]
    joined = btc.join(
        eth,
        on=["ts", "pillar", "family", "component"],
        how="inner",
        suffix="_eth",
    )
    for item in joined.iter_rows(named=True):
        rows.append(
            {
                "ts": item["ts"],
                "date": item["ts"].date(),
                "scope": MARKET_SCOPE,
                "asset": MARKET_ASSET,
                "source": source,
                "pillar": item["pillar"],
                "family": item["family"],
                "component": item["component"],
                "raw_value": _weighted_pair(item["raw_value"], item["raw_value_eth"]),
                "normalized_value": _weighted_pair(
                    item["normalized_value"], item["normalized_value_eth"]
                ),
                "component_weight": _weighted_pair(
                    item["component_weight"], item["component_weight_eth"]
                ),
                "pillar_score": _weighted_pair(item["pillar_score"], item["pillar_score_eth"]),
                "pillar_weight": _weighted_pair(item["pillar_weight"], item["pillar_weight_eth"]),
                "signed_contribution": _weighted_pair(
                    item["signed_contribution"], item["signed_contribution_eth"]
                ),
                "quality_flag": _merge_quality_flags(
                    str(item["quality_flag"]), str(item["quality_flag_eth"])
                ),
            }
        )
    return _frame_from_rows(DatasetName.COMPOSITE_COMPONENT_SERIES, rows)


def build_composite_transition_series(
    store: DatasetStore,
    start: datetime,
    end: datetime,
    *,
    asset: str,
    source: str = SOURCE,
) -> pl.DataFrame:
    frame = store.read(
        DatasetName.COMPOSITE_SIGNAL_SERIES,
        start=start - timedelta(hours=CALIBRATION_WINDOW_HOURS),
        end=end,
        filters={"asset": asset.upper(), "scope": ASSET_SCOPE, "source": source},
    )
    return _build_transition_frame(frame, scope=ASSET_SCOPE, asset=asset.upper(), source=source)


def build_market_composite_transition_series(
    store: DatasetStore,
    start: datetime,
    end: datetime,
    *,
    source: str = SOURCE,
) -> pl.DataFrame:
    frame = store.read(
        DatasetName.COMPOSITE_SIGNAL_SERIES,
        start=start - timedelta(hours=CALIBRATION_WINDOW_HOURS),
        end=end,
        filters={"asset": MARKET_ASSET, "scope": MARKET_SCOPE, "source": source},
    )
    return _build_transition_frame(frame, scope=MARKET_SCOPE, asset=MARKET_ASSET, source=source)


def _build_asset_raw_frame(
    store: DatasetStore,
    start: datetime,
    end: datetime,
    asset: str,
) -> pl.DataFrame:
    options = _load_asof_frame(
        store.read(
            DatasetName.OPTIONS_SIGNAL_SERIES,
            start=start,
            end=end,
            filters={"underlying": asset, "source": "options_composite"},
        ),
        "options",
        timedelta(hours=12),
        [
            "vol_carry_score",
            "skew_stress_score",
            "ml_carry_toxicity_score",
            "funding_skew_score",
            "sessionality_score",
            "dispersion_score",
            "expiry_pin_score",
            "gamma_flip_score",
            "signal_state",
            "quality_flag",
        ],
    )
    vvix = _load_asof_frame(
        store.read(
            DatasetName.VVIX_SERIES,
            start=start,
            end=end,
            filters={"underlying": asset, "source": "dvol"},
        ),
        "vvix",
        timedelta(hours=12),
        ["rvvix_14d", "qvvix_30d", "quality_flag"],
    )
    razor = _load_asof_frame(
        store.read(
            DatasetName.RAZOR_SERIES,
            start=start,
            end=end,
            filters={"underlying": asset, "source": "synthetic_7d_atm"},
        ),
        "razor",
        timedelta(hours=12),
        ["razor_pnl_zscore", "razor_vrp_zscore", "quality_flag"],
    )
    perps_signal = _load_asof_frame(
        store.read(
            DatasetName.PERPS_SIGNAL_SERIES,
            start=start,
            end=end,
            filters={"asset": asset, "venue_scope": "all", "source": "perps_family_v1"},
        ),
        "perps_signal",
        timedelta(hours=6),
        [
            "carry_momentum_1h",
            "carry_momentum_24h",
            "oi_price_divergence",
            "cum_delta_acceleration",
            "delta_divergence",
            "lead_lag_score",
            "depth_recovery_speed",
            "depth_imbalance_momentum",
            "spread_zscore_4h",
            "oi_concentration",
            "quality_flag",
        ],
    )
    perps_composite = _load_asof_frame(
        store.read(
            DatasetName.PERPS_COMPOSITE_SERIES,
            start=start,
            end=end,
            filters={"asset": asset, "venue_scope": "all", "source": "perps_composite_v1"},
        ),
        "perps_composite",
        timedelta(hours=6),
        [
            "carry_score",
            "cascade_risk_score",
            "toxicity_score",
            "positioning_score",
            "fragility_score",
            "quality_flag",
        ],
    )
    micro = _build_micro_hourly_frame(store, start, end, asset)
    price = _build_hourly_price_frame(store, start, end, asset)

    timelines = [
        frame.select(pl.col(frame.columns[0]).alias("ts"))
        for frame in (options, vvix, razor, perps_signal, perps_composite, micro, price)
        if not frame.is_empty()
    ]
    if not timelines:
        return pl.DataFrame()
    timeline = pl.concat(timelines, how="vertical").unique().sort("ts")
    frame = timeline
    frame = _join_source(frame, options, "options_source_ts")
    frame = _join_source(frame, vvix, "vvix_source_ts")
    frame = _join_source(frame, razor, "razor_source_ts")
    frame = _join_source(frame, perps_signal, "perps_signal_source_ts")
    frame = _join_source(frame, perps_composite, "perps_composite_source_ts")
    frame = _join_source(frame, micro, "micro_source_ts")
    frame = _join_source(frame, price, "price_source_ts")
    return frame.sort("ts")


def _build_micro_hourly_frame(
    store: DatasetStore,
    start: datetime,
    end: datetime,
    asset: str,
) -> pl.DataFrame:
    frame = store.read(
        DatasetName.MICROSTRUCTURE_MODEL_OUTPUT,
        start=start,
        end=end,
        filters={"asset": asset, "source": "tcn_boost_fusion"},
    )
    if frame.is_empty():
        return pl.DataFrame()
    horizon_weight_expr = (
        pl.when(pl.col("horizon") == "30s")
        .then(pl.lit(HORIZON_WEIGHTS["30s"]))
        .when(pl.col("horizon") == "1m")
        .then(pl.lit(HORIZON_WEIGHTS["1m"]))
        .when(pl.col("horizon") == "2m")
        .then(pl.lit(HORIZON_WEIGHTS["2m"]))
        .otherwise(pl.lit(HORIZON_WEIGHTS["5m"]))
    )
    weighted = frame.with_columns(
        [
            pl.col("ts").dt.truncate("1h").alias("micro_source_ts"),
            horizon_weight_expr.alias("horizon_weight"),
            ((pl.col("prob_up") - pl.col("prob_down")) * pl.col("confidence")).alias(
                "signed_score"
            ),
            pl.when(pl.col("blocked_reason").is_null())
            .then(0.0)
            .otherwise(1.0)
            .alias("blocked_flag"),
        ]
    )
    grouped = weighted.group_by("micro_source_ts").agg(
        [
            (
                (pl.col("signed_score") * pl.col("horizon_weight")).sum()
                / pl.col("horizon_weight").sum()
            ).alias("micro_direction_raw"),
            (pl.mean("signed_score") * pl.mean("tradability_score")).alias("micro_sentiment_raw"),
            pl.mean("confidence").alias("micro_confidence"),
            pl.mean("blocked_flag").alias("micro_blocked_rate"),
            pl.std("signed_score").alias("micro_horizon_dispersion"),
            pl.n_unique("horizon").alias("micro_active_horizons"),
            pl.max("quality_flag").alias("micro_quality_flag"),
        ]
    )
    return grouped.with_columns(
        [
            (
                (1.0 - pl.col("micro_confidence").fill_null(0.0))
                + pl.col("micro_blocked_rate").fill_null(0.0)
                + pl.col("micro_horizon_dispersion").fill_null(0.0).clip(0.0, 1.0)
            ).alias("micro_fragility_raw"),
            pl.lit(timedelta(hours=2)).alias("_tolerance"),
        ]
    )


def _build_hourly_price_frame(
    store: DatasetStore,
    start: datetime,
    end: datetime,
    asset: str,
) -> pl.DataFrame:
    bars = store.read(
        DatasetName.MICROSTRUCTURE_BAR,
        start=start,
        end=end + timedelta(hours=2),
        filters={"asset": asset, "bar_size_secs": 60},
    )
    if bars.is_empty():
        return pl.DataFrame()
    preferred = bars.filter(pl.col("market_type") == "spot")
    if preferred.is_empty():
        preferred = bars
    hourly = (
        preferred.with_columns(pl.col("ts").dt.truncate("1h").alias("price_source_ts"))
        .sort(["price_source_ts", "ts"])
        .group_by("price_source_ts")
        .agg(pl.col("close").last().alias("hourly_close"))
        .sort("price_source_ts")
        .with_columns(
            (((pl.col("hourly_close").shift(-1) / pl.col("hourly_close")) - 1.0) * 10_000.0).alias(
                "next_return_bps"
            )
        )
        .with_columns(pl.lit(timedelta(hours=2)).alias("_tolerance"))
    )
    return hourly


def _load_asof_frame(
    frame: pl.DataFrame,
    prefix: str,
    tolerance: timedelta,
    columns: list[str],
) -> pl.DataFrame:
    if frame.is_empty():
        return pl.DataFrame()
    selected = ["ts", *[column for column in columns if column in frame.columns]]
    source_ts_column = f"{prefix}_source_ts"
    renamed = frame.select(selected).rename({"ts": source_ts_column})
    rename_map = {
        column: f"{prefix}_{column}" for column in renamed.columns if column != source_ts_column
    }
    return renamed.rename(rename_map).with_columns(pl.lit(tolerance).alias("_tolerance"))


def _normalize_components(frame: pl.DataFrame) -> pl.DataFrame:
    result = frame.sort("ts").with_row_index("_history_points", offset=1)
    for spec in COMPONENT_SPECS:
        if spec.source_column not in result.columns:
            result = result.with_columns(pl.lit(None, dtype=pl.Float64).alias(spec.source_column))
        result = result.with_columns(
            _normalized_expr(spec.source_column, spec.sign).alias(f"norm__{spec.component}")
        )
    return result


def _normalized_expr(column: str, sign: float) -> pl.Expr:
    primary = (
        (
            pl.col(column)
            - pl.col(column).rolling_mean(
                NORMALIZATION_WINDOW_HOURS, min_samples=NORMALIZATION_MIN_SAMPLES
            )
        )
        / (
            pl.col(column).rolling_std(
                NORMALIZATION_WINDOW_HOURS, min_samples=NORMALIZATION_MIN_SAMPLES
            )
            + 1e-6
        )
    )
    fallback = (
        (
            pl.col(column)
            - pl.col(column).rolling_mean(
                NORMALIZATION_WINDOW_HOURS, min_samples=NORMALIZATION_FALLBACK_MIN_SAMPLES
            )
        )
        / (
            pl.col(column).rolling_std(
                NORMALIZATION_WINDOW_HOURS,
                min_samples=NORMALIZATION_FALLBACK_MIN_SAMPLES,
            )
            + 1e-6
        )
    )
    base = pl.when(primary.is_not_null()).then(primary).otherwise(fallback * 0.6)
    base = base.clip(-3.0, 3.0) / 3.0
    if sign < 0:
        base = -base
    return base


def _build_component_frame(
    frame: pl.DataFrame,
    *,
    asset: str,
    scope: str,
    source: str,
) -> pl.DataFrame:
    rows: list[dict[str, Any]] = []
    for item in frame.iter_rows(named=True):
        family_scores = _compute_family_scores(item)
        pillar_scores = _compute_pillar_scores(family_scores)
        for spec in COMPONENT_SPECS:
            normalized_value = item.get(f"norm__{spec.component}")
            if normalized_value is None:
                continue
            family_total = _family_total_weight(spec.pillar, spec.family)
            family_weight = spec.weight / family_total if family_total else 0.0
            pillar_weight = PILLAR_WEIGHTS[spec.pillar][spec.family]
            signed_contribution = float(normalized_value) * family_weight * pillar_weight
            rows.append(
                {
                    "ts": item["ts"],
                    "date": item["ts"].date(),
                    "scope": scope,
                    "asset": asset,
                    "source": source,
                    "pillar": spec.pillar,
                    "family": spec.family,
                    "component": spec.component,
                    "raw_value": item.get(spec.source_column),
                    "normalized_value": float(normalized_value),
                    "component_weight": family_weight,
                    "pillar_score": pillar_scores.get(spec.pillar),
                    "pillar_weight": pillar_weight,
                    "signed_contribution": signed_contribution,
                    "quality_flag": _component_quality(item, spec),
                }
            )
    if not rows:
        return empty_frame(DatasetName.COMPOSITE_COMPONENT_SERIES)
    return _frame_from_rows(DatasetName.COMPOSITE_COMPONENT_SERIES, rows)


def _build_signal_frame(
    frame: pl.DataFrame,
    component_frame: pl.DataFrame,
    *,
    asset: str,
    scope: str,
    source: str,
) -> pl.DataFrame:
    rows: list[dict[str, Any]] = []
    for item in frame.iter_rows(named=True):
        family_scores = _compute_family_scores(item)
        coverage_ratio = _coverage_ratio(item)
        freshness_score = _freshness_score(item)
        confidence = _confidence_score(item, coverage_ratio, freshness_score)
        sentiment_raw = _weighted_family_value(
            family_scores,
            "sentiment",
            {
                "options": "options_sentiment_subindex",
                "perps": "perps_sentiment_subindex",
                "micro": "micro_sentiment_subindex",
            },
        )
        directional_raw = _weighted_family_value(
            family_scores,
            "directional",
            {
                "micro": "micro_direction_subindex",
                "perps": "perps_direction_subindex",
                "options": "options_direction_subindex",
            },
        )
        fragility_raw = _weighted_family_value(
            family_scores,
            "fragility",
            {
                "perps": "perps_fragility_subindex",
                "options": "options_fragility_subindex",
                "micro": "micro_fragility_subindex",
            },
        )
        rows.append(
            {
                "ts": item["ts"],
                "date": item["ts"].date(),
                "scope": scope,
                "asset": asset,
                "source": source,
                "sentiment_index": _sentiment_fallback(sentiment_raw),
                "directional_tilt": _directional_fallback(directional_raw),
                "fragility_index": _fragility_fallback(fragility_raw),
                "confidence_score": confidence,
                "options_sentiment_subindex": family_scores.get("options_sentiment_subindex"),
                "perps_sentiment_subindex": family_scores.get("perps_sentiment_subindex"),
                "micro_sentiment_subindex": family_scores.get("micro_sentiment_subindex"),
                "micro_direction_subindex": family_scores.get("micro_direction_subindex"),
                "perps_direction_subindex": family_scores.get("perps_direction_subindex"),
                "options_direction_subindex": family_scores.get("options_direction_subindex"),
                "perps_fragility_subindex": family_scores.get("perps_fragility_subindex"),
                "options_fragility_subindex": family_scores.get("options_fragility_subindex"),
                "micro_fragility_subindex": family_scores.get("micro_fragility_subindex"),
                "coverage_ratio": coverage_ratio,
                "freshness_score": freshness_score,
                "calibrator_mode": "fixed_fallback",
                "sentiment_regime": _sentiment_regime(_sentiment_fallback(sentiment_raw)),
                "directional_regime": _directional_regime(_directional_fallback(directional_raw)),
                "fragility_regime": _fragility_regime(_fragility_fallback(fragility_raw)),
                "quality_flag": _quality_flag(coverage_ratio, confidence),
            }
        )
    if not rows:
        return empty_frame(DatasetName.COMPOSITE_SIGNAL_SERIES)
    return _frame_from_rows(DatasetName.COMPOSITE_SIGNAL_SERIES, rows)


def _compute_family_scores(item: dict[str, Any]) -> dict[str, float | None]:
    scores: dict[str, float | None] = {}
    for pillar in PILLAR_WEIGHTS:
        for family in PILLAR_WEIGHTS[pillar]:
            subset = [
                spec for spec in COMPONENT_SPECS if spec.pillar == pillar and spec.family == family
            ]
            total_weight = 0.0
            weighted_sum = 0.0
            for spec in subset:
                value = item.get(f"norm__{spec.component}")
                if value is None:
                    continue
                weighted_sum += float(value) * spec.weight
                total_weight += spec.weight
            key = (
                f"{family}_{pillar}_subindex"
                if pillar != "directional" or family != "micro"
                else "micro_direction_subindex"
            )
            if total_weight:
                scores[key] = weighted_sum / total_weight
            else:
                scores[key] = None
    # normalize names to the dataset schema we expose
    return {
        "options_sentiment_subindex": scores.get("options_sentiment_subindex"),
        "perps_sentiment_subindex": scores.get("perps_sentiment_subindex"),
        "micro_sentiment_subindex": scores.get("micro_sentiment_subindex"),
        "micro_direction_subindex": scores.get("micro_direction_subindex"),
        "perps_direction_subindex": scores.get("perps_direction_subindex"),
        "options_direction_subindex": scores.get("options_direction_subindex"),
        "perps_fragility_subindex": scores.get("perps_fragility_subindex"),
        "options_fragility_subindex": scores.get("options_fragility_subindex"),
        "micro_fragility_subindex": scores.get("micro_fragility_subindex"),
    }


def _compute_pillar_scores(family_scores: dict[str, float | None]) -> dict[str, float | None]:
    return {
        "sentiment": _weighted_mean(
            [
                (
                    family_scores.get("options_sentiment_subindex"),
                    PILLAR_WEIGHTS["sentiment"]["options"],
                ),
                (
                    family_scores.get("perps_sentiment_subindex"),
                    PILLAR_WEIGHTS["sentiment"]["perps"],
                ),
                (
                    family_scores.get("micro_sentiment_subindex"),
                    PILLAR_WEIGHTS["sentiment"]["micro"],
                ),
            ]
        ),
        "directional": _weighted_mean(
            [
                (
                    family_scores.get("micro_direction_subindex"),
                    PILLAR_WEIGHTS["directional"]["micro"],
                ),
                (
                    family_scores.get("perps_direction_subindex"),
                    PILLAR_WEIGHTS["directional"]["perps"],
                ),
                (
                    family_scores.get("options_direction_subindex"),
                    PILLAR_WEIGHTS["directional"]["options"],
                ),
            ]
        ),
        "fragility": _weighted_mean(
            [
                (
                    family_scores.get("perps_fragility_subindex"),
                    PILLAR_WEIGHTS["fragility"]["perps"],
                ),
                (
                    family_scores.get("options_fragility_subindex"),
                    PILLAR_WEIGHTS["fragility"]["options"],
                ),
                (
                    family_scores.get("micro_fragility_subindex"),
                    PILLAR_WEIGHTS["fragility"]["micro"],
                ),
            ]
        ),
    }


def _apply_calibration(signal_frame: pl.DataFrame, raw_frame: pl.DataFrame) -> pl.DataFrame:
    if signal_frame.is_empty():
        return signal_frame
    calibrated = signal_frame.sort("ts")
    calibrated = _apply_percentile_calibration(
        calibrated,
        "sentiment_index",
        "options_sentiment_subindex",
        "perps_sentiment_subindex",
        "micro_sentiment_subindex",
    )
    calibrated = _apply_percentile_calibration(
        calibrated,
        "fragility_index",
        "perps_fragility_subindex",
        "options_fragility_subindex",
        "micro_fragility_subindex",
    )
    calibrated = _apply_directional_calibration(calibrated, raw_frame)
    return calibrated.with_columns(
        [
            pl.col("sentiment_index")
            .map_elements(_sentiment_regime, return_dtype=pl.String)
            .alias("sentiment_regime"),
            pl.col("directional_tilt")
            .map_elements(_directional_regime, return_dtype=pl.String)
            .alias("directional_regime"),
            pl.col("fragility_index")
            .map_elements(_fragility_regime, return_dtype=pl.String)
            .alias("fragility_regime"),
            pl.struct(["coverage_ratio", "confidence_score"])
            .map_elements(
                lambda item: _quality_flag(item["coverage_ratio"], item["confidence_score"]),
                return_dtype=pl.String,
            )
            .alias("quality_flag"),
        ]
    )


def _apply_percentile_calibration(
    frame: pl.DataFrame,
    target_column: str,
    *support_columns: str,
) -> pl.DataFrame:
    values = frame.get_column(target_column).to_list()
    valid = [float(value) for value in values if value is not None]
    if len(valid) < CALIBRATION_MIN_SAMPLES:
        return frame
    ranked = (
        pl.DataFrame({"value": valid})
        .with_row_index("row_nr")
        .with_columns(
            (pl.col("value").rank(method="average").truediv(len(valid)).mul(100.0)).alias(
                "percentile"
            )
        )
        .sort("row_nr")
    )
    x = ranked.get_column("value").to_numpy()
    y = ranked.get_column("percentile").to_numpy()
    model = IsotonicRegression(y_min=0.0, y_max=100.0, out_of_bounds="clip")
    model.fit(x, y)
    calibrated_values: list[float | None] = []
    valid_iter = iter(model.predict(np.array(valid, dtype=float)).tolist())
    for value in values:
        calibrated_values.append(None if value is None else float(next(valid_iter)))
    mode = "isotonic_percentile"
    return frame.with_columns(
        [
            pl.Series(name=target_column, values=calibrated_values, dtype=pl.Float64),
            pl.when(pl.all_horizontal([pl.col(column).is_null() for column in support_columns]))
            .then(pl.col("calibrator_mode"))
            .otherwise(pl.lit(mode))
            .alias("calibrator_mode"),
        ]
    )


def _apply_directional_calibration(
    signal_frame: pl.DataFrame, raw_frame: pl.DataFrame
) -> pl.DataFrame:
    if "next_return_bps" not in raw_frame.columns:
        return signal_frame
    calibration_columns: list[pl.Expr] = [
        pl.col("ts"),
        pl.col("next_return_bps"),
        (
            pl.col("micro_confidence")
            if "micro_confidence" in raw_frame.columns
            else pl.lit(None, dtype=pl.Float64).alias("micro_confidence")
        ),
        (
            pl.col("micro_active_horizons")
            if "micro_active_horizons" in raw_frame.columns
            else pl.lit(None, dtype=pl.Float64).alias("micro_active_horizons")
        ),
    ]
    joined = signal_frame.join(
        raw_frame.select(calibration_columns),
        on="ts",
        how="left",
    )
    valid = joined.filter(
        pl.col("directional_tilt").is_not_null()
        & pl.col("next_return_bps").is_not_null()
        & pl.col("coverage_ratio").is_not_null()
    )
    if valid.height < CALIBRATION_MIN_SAMPLES:
        return signal_frame
    x = valid.select(
        [
            ((pl.col("directional_tilt") / 100.0).alias("raw_direction")),
            pl.col("coverage_ratio"),
            pl.col("micro_confidence").fill_null(0.0),
            (pl.col("micro_active_horizons").fill_null(0.0) / len(HORIZON_WEIGHTS)).alias(
                "horizon_coverage"
            ),
        ]
    ).to_numpy()
    y = np.clip(valid.get_column("next_return_bps").to_numpy(), -50.0, 50.0)
    model = Ridge(alpha=1.0)
    model.fit(x, y)
    full_x = joined.select(
        [
            ((pl.col("directional_tilt") / 100.0).fill_null(0.0).alias("raw_direction")),
            pl.col("coverage_ratio").fill_null(0.0),
            pl.col("micro_confidence").fill_null(0.0),
            (pl.col("micro_active_horizons").fill_null(0.0) / len(HORIZON_WEIGHTS)).alias(
                "horizon_coverage"
            ),
        ]
    ).to_numpy()
    predictions = model.predict(full_x)
    tilt = [float(np.clip(np.tanh(pred / 12.0) * 100.0, -100.0, 100.0)) for pred in predictions]
    return signal_frame.with_columns(
        [
            pl.Series(name="directional_tilt", values=tilt, dtype=pl.Float64),
            pl.lit("ridge_directional").alias("calibrator_mode"),
        ]
    )


def _build_transition_frame(
    frame: pl.DataFrame,
    *,
    scope: str,
    asset: str,
    source: str,
) -> pl.DataFrame:
    if frame.is_empty():
        return empty_frame(DatasetName.COMPOSITE_TRANSITION_SERIES)
    rows: list[dict[str, Any]] = []
    ordered = frame.sort("ts")
    for index_type, column in (
        ("sentiment", "sentiment_regime"),
        ("directional", "directional_regime"),
        ("fragility", "fragility_regime"),
    ):
        regimes = ordered.select(["ts", column]).drop_nulls()
        if regimes.is_empty():
            continue
        order = REGIME_ORDER[index_type]
        values = regimes.get_column(column).to_list()
        timestamps = regimes.get_column("ts").to_list()
        durations = _regime_durations(values, timestamps)
        for idx, (ts, current) in enumerate(zip(timestamps, values, strict=False)):
            prior = values[idx - 1] if idx else None
            window_values = values[max(0, idx - CALIBRATION_WINDOW_HOURS) : idx + 1]
            transitions = list(zip(window_values[:-1], window_values[1:], strict=False))
            outgoing = [nxt for prev, nxt in transitions if prev == current]
            stay_probability = _probability(outgoing, current)
            current_rank = order.index(current) if current in order else None
            upshift_probability = None
            downshift_probability = None
            if current_rank is not None and outgoing:
                upshift_probability = float(
                    sum(1 for nxt in outgoing if nxt in order and order.index(nxt) > current_rank)
                    / len(outgoing)
                )
                downshift_probability = float(
                    sum(1 for nxt in outgoing if nxt in order and order.index(nxt) < current_rank)
                    / len(outgoing)
                )
            rows.append(
                {
                    "ts": ts,
                    "date": ts.date(),
                    "scope": scope,
                    "asset": asset,
                    "source": source,
                    "index_type": index_type,
                    "current_regime": current,
                    "prior_regime": prior,
                    "stay_probability": stay_probability,
                    "upshift_probability": upshift_probability,
                    "downshift_probability": downshift_probability,
                    "median_regime_duration_hours": durations.get(current),
                    "quality_flag": "ok"
                    if len(window_values) >= NORMALIZATION_MIN_SAMPLES
                    else "partial_context",
                }
            )
    return _frame_from_rows(DatasetName.COMPOSITE_TRANSITION_SERIES, rows)


def _family_total_weight(pillar: str, family: str) -> float:
    return sum(
        spec.weight for spec in COMPONENT_SPECS if spec.pillar == pillar and spec.family == family
    )


def _coverage_ratio(item: dict[str, Any]) -> float:
    total = len(COMPONENT_SPECS)
    available = sum(
        1 for spec in COMPONENT_SPECS if item.get(f"norm__{spec.component}") is not None
    )
    return float(available / total) if total else 0.0


def _freshness_score(item: dict[str, Any]) -> float:
    ages = []
    ts = item["ts"]
    for prefix in ("options", "vvix", "razor", "perps_signal", "perps_composite", "micro", "price"):
        source_ts = item.get(f"{prefix}_source_ts")
        if source_ts is None:
            continue
        age_hours = max((ts - source_ts).total_seconds() / 3600.0, 0.0)
        ages.append(max(0.0, 1.0 - min(age_hours / 6.0, 1.0)))
    return float(np.mean(ages)) if ages else 0.0


def _confidence_score(item: dict[str, Any], coverage_ratio: float, freshness_score: float) -> float:
    quality_penalty = 0.0
    for column in (
        "options_quality_flag",
        "vvix_quality_flag",
        "razor_quality_flag",
        "perps_signal_quality_flag",
        "perps_composite_quality_flag",
        "micro_quality_flag",
    ):
        value = item.get(column)
        if value in {"partial_context", "gated"}:
            quality_penalty += 0.05
        elif value in {"insufficient_data", "blocked"}:
            quality_penalty += 0.15
    micro_confidence = float(item.get("micro_confidence") or 0.0)
    history_points = float(item.get("_history_points") or 0.0)
    history_score = min(history_points / NORMALIZATION_MIN_SAMPLES, 1.0)
    score = (
        0.30 * coverage_ratio
        + 0.20 * freshness_score
        + 0.25 * micro_confidence
        + 0.25 * history_score
        - quality_penalty
    )
    return float(np.clip(score, 0.0, 1.0))


def _weighted_family_value(
    family_scores: dict[str, float | None],
    pillar: str,
    mapping: dict[str, str],
) -> float | None:
    pairs = [
        (family_scores.get(key), PILLAR_WEIGHTS[pillar][family]) for family, key in mapping.items()
    ]
    return _weighted_mean(pairs)


def _weighted_mean(values: list[tuple[float | None, float]]) -> float | None:
    numerator = 0.0
    denominator = 0.0
    for value, weight in values:
        if value is None:
            continue
        numerator += float(value) * weight
        denominator += weight
    if denominator == 0.0:
        return None
    return numerator / denominator


def _sentiment_fallback(value: float | None) -> float | None:
    if value is None:
        return None
    return float(np.clip(50.0 + 25.0 * np.tanh(value * 1.5), 0.0, 100.0))


def _directional_fallback(value: float | None) -> float | None:
    if value is None:
        return None
    return float(np.clip(np.tanh(value * 1.5) * 100.0, -100.0, 100.0))


def _fragility_fallback(value: float | None) -> float | None:
    if value is None:
        return None
    return float(np.clip(50.0 + 25.0 * np.tanh(value * 1.5), 0.0, 100.0))


def _sentiment_regime(value: float | None) -> str | None:
    if value is None:
        return None
    for threshold, label in SENTIMENT_REGIMES:
        if value < threshold:
            return label
    return "euphoria"


def _directional_regime(value: float | None) -> str | None:
    if value is None:
        return None
    if value <= -40.0:
        return "bearish"
    if value <= -15.0:
        return "mild_bearish"
    if value < 15.0:
        return "neutral"
    if value < 40.0:
        return "mild_bullish"
    return "bullish"


def _fragility_regime(value: float | None) -> str | None:
    if value is None:
        return None
    for threshold, label in FRAGILITY_REGIMES:
        if value < threshold:
            return label
    return "squeeze_risk"


def _quality_flag(coverage_ratio: float | None, confidence: float | None) -> str:
    if coverage_ratio is None or coverage_ratio < 0.2:
        return "insufficient_data"
    if confidence is None or confidence < 0.4 or coverage_ratio < 0.75:
        return "partial_context"
    return "ok"


def _component_quality(item: dict[str, Any], spec: ComponentSpec) -> str:
    value = item.get(f"norm__{spec.component}")
    if value is None:
        return "insufficient_data"
    return "ok"


def _join_market_frames(frames: dict[str, pl.DataFrame]) -> pl.DataFrame:
    joined = frames["BTC"].rename(
        {column: f"{column}_btc" for column in frames["BTC"].columns if column != "ts"}
    )
    joined = joined.join(
        frames["ETH"].rename(
            {column: f"{column}_eth" for column in frames["ETH"].columns if column != "ts"}
        ),
        on="ts",
        how="inner",
    )
    return joined


def _weighted_market_value(item: dict[str, Any], field: str) -> float | None:
    return _weighted_pair(item.get(f"{field}_btc"), item.get(f"{field}_eth"))


def _weighted_pair(first: Any, second: Any) -> float | None:
    pairs = []
    if first is not None:
        pairs.append((float(first), MARKET_WEIGHTS["BTC"]))
    if second is not None:
        pairs.append((float(second), MARKET_WEIGHTS["ETH"]))
    if not pairs:
        return None
    return float(
        sum(value * weight for value, weight in pairs) / sum(weight for _, weight in pairs)
    )


def _merge_quality_flags(first: str | None, second: str | None) -> str:
    values = {first, second}
    if "insufficient_data" in values:
        return "insufficient_data"
    if "partial_context" in values:
        return "partial_context"
    return "ok"


def _probability(values: list[str], target: str) -> float | None:
    if not values:
        return None
    return float(sum(1 for value in values if value == target) / len(values))


def _regime_durations(regimes: list[str], timestamps: list[datetime]) -> dict[str, float]:
    durations: dict[str, list[float]] = {}
    if not regimes:
        return {}
    current = regimes[0]
    start = timestamps[0]
    for regime, ts in zip(regimes[1:], timestamps[1:], strict=False):
        if regime != current:
            durations.setdefault(current, []).append((ts - start).total_seconds() / 3600.0)
            current = regime
            start = ts
    durations.setdefault(current, []).append((timestamps[-1] - start).total_seconds() / 3600.0)
    return {regime: float(np.median(values)) for regime, values in durations.items()}


def _join_source(base: pl.DataFrame, source_frame: pl.DataFrame, ts_column: str) -> pl.DataFrame:
    if source_frame.is_empty():
        return base
    tolerance = source_frame.item(0, "_tolerance")
    return base.join_asof(
        source_frame.sort(ts_column),
        left_on="ts",
        right_on=ts_column,
        strategy="backward",
        tolerance=tolerance,
    ).drop("_tolerance")



from __future__ import annotations

from fastapi.testclient import TestClient

from uforia.api.app import create_app


def test_api_allows_private_lan_origin_for_browser_refresh() -> None:
    client = TestClient(create_app())

    response = client.options(
        "/api/v1/system/health",
        headers={
            "Origin": "http://192.168.1.69:3000",
            "Access-Control-Request-Method": "GET",
        },
    )

    assert response.status_code == 200
    assert response.headers["access-control-allow-origin"] == "http://192.168.1.69:3000"



from __future__ import annotations

from datetime import datetime, timezone

import polars as pl
from fastapi.testclient import TestClient

from uforia.api.app import create_app
from uforia.api.settings import ApiSettings
from uforia.storage import DatasetName, DatasetStore


def test_api_exposes_perps_dashboards_and_signals(tmp_path) -> None:
    store = DatasetStore(tmp_path)
    base = datetime(2026, 3, 1, 12, tzinfo=timezone.utc)
    store.write(
        DatasetName.PERPS_SIGNAL_SERIES,
        pl.DataFrame(
            [
                {
                    "ts": base,
                    "date": base.date(),
                    "asset": "BTC",
                    "venue_scope": "all",
                    "source": "perps_family_v1",
                    "basis_bps": 12.5,
                    "funding_rate_1h": 0.0000015,
                    "carry_annualized_pct": 18.2,
                    "carry_zscore_7d": 1.4,
                    "carry_zscore_30d": 1.1,
                    "carry_momentum_1h": 0.3,
                    "carry_momentum_24h": 1.2,
                    "cross_venue_funding_spread": 0.0000003,
                    "long_liq_notional": 500_000.0,
                    "short_liq_notional": 350_000.0,
                    "liq_intensity_zscore": 1.7,
                    "liq_asymmetry": 0.18,
                    "oi_concentration": 0.42,
                    "crowding_score": 1.9,
                    "cascade_risk_score": 2.1,
                    "cascade_regime": "high",
                    "vpin_50": 0.61,
                    "toxic_flow_zscore_4h": 1.2,
                    "toxic_flow_zscore_24h": 0.8,
                    "ofi_persistence": 0.35,
                    "large_trade_ratio": 0.16,
                    "toxicity_regime": "elevated",
                    "oi_change_1h": 3.1,
                    "oi_change_4h": 7.2,
                    "oi_change_24h": 11.5,
                    "price_change_1h": 24.0,
                    "oi_price_divergence": 3.1,
                    "oi_acceleration": 0.8,
                    "leverage_regime": "building",
                    "lead_lag_score": 0.0003,
                    "venue_dislocation_bps": 1.4,
                    "convergence_half_life_secs": 145.0,
                    "price_discovery_leader": "binance",
                    "depth_recovery_speed": 0.8,
                    "spread_bps": 2.3,
                    "spread_zscore_4h": 0.9,
                    "spread_zscore_24h": 1.2,
                    "depth_imbalance_momentum": 0.05,
                    "spread_regime": "normal",
                    "rv_5s": 0.61,
                    "rv_1m": 0.54,
                    "rv_5m": 0.48,
                    "vol_signature_ratio": 1.27,
                    "vol_clustering": 0.36,
                    "rv_regime": "elevated",
                    "cum_delta_acceleration": 15.0,
                    "delta_divergence": 4.0,
                    "absorption_ratio": 750.0,
                    "delta_regime": "aggressive_buying",
                    "quality_flag": "ok",
                }
            ]
        ),
    )
    store.write(
        DatasetName.PERPS_COMPOSITE_SERIES,
        pl.DataFrame(
            [
                {
                    "ts": base,
                    "date": base.date(),
                    "asset": "BTC",
                    "venue_scope": "all",
                    "source": "perps_composite_v1",
                    "carry_score": 1.25,
                    "cascade_risk_score": 2.1,
                    "toxicity_score": 1.05,
                    "positioning_score": 0.88,
                    "fragility_score": 0.66,
                    "quality_flag": "ok",
                }
            ]
        ),
    )

    app = create_app(settings=ApiSettings(data_root=tmp_path), store=store)
    client = TestClient(app)

    dashboards = client.get("/api/v1/perps/dashboards")
    dashboard_series = client.get(
        "/api/v1/perps/funding_basis_carry/series",
        params={"asset": "BTC", "venue_scope": "all"},
    )
    signals = client.get("/api/v1/perps/signals")
    signal_latest = client.get(
        "/api/v1/perps/signals/carry_score/latest",
        params={"asset": "BTC", "venue_scope": "all"},
    )

    assert dashboards.status_code == 200
    assert {
        "funding_basis_carry",
        "liquidation_cascade",
        "order_flow_toxicity",
        "oi_price_divergence",
        "cross_venue_price_discovery",
        "depth_resilience",
        "vol_microstructure",
        "cum_delta_momentum",
    } <= {item["id"] for item in dashboards.json()["data"]}
    assert dashboard_series.status_code == 200
    assert len(dashboard_series.json()["data"]["points"]) == 1
    assert signals.status_code == 200
    assert {
        "carry_score",
        "cascade_risk_score",
        "toxicity_score",
        "positioning_score",
        "fragility_score",
    } <= {item["id"] for item in signals.json()["data"]}
    assert signal_latest.status_code == 200
    assert signal_latest.json()["data"]["latest_values"]["carry_score"] == 1.25


def test_perps_dashboard_quality_uses_dashboard_specific_coverage(tmp_path) -> None:
    store = DatasetStore(tmp_path)
    base = datetime(2026, 3, 1, 12, tzinfo=timezone.utc)
    row = {
        "ts": base,
        "date": base.date(),
        "asset": "BTC",
        "venue_scope": "all",
        "source": "perps_family_v1",
        "basis_bps": 12.5,
        "funding_rate_1h": 0.0000015,
        "carry_annualized_pct": 18.2,
        "carry_zscore_7d": 1.4,
        "carry_zscore_30d": 1.1,
        "carry_momentum_1h": 0.3,
        "carry_momentum_24h": 1.2,
        "cross_venue_funding_spread": None,
        "long_liq_notional": 500_000.0,
        "short_liq_notional": 350_000.0,
        "liq_intensity_zscore": 1.7,
        "liq_asymmetry": 0.18,
        "oi_concentration": 0.42,
        "crowding_score": 1.9,
        "cascade_risk_score": 2.1,
        "cascade_regime": "high",
        "vpin_50": 0.61,
        "toxic_flow_zscore_4h": 1.2,
        "toxic_flow_zscore_24h": None,
        "ofi_persistence": 0.35,
        "large_trade_ratio": 0.16,
        "toxicity_regime": "elevated",
        "oi_change_1h": 3.1,
        "oi_change_4h": 7.2,
        "oi_change_24h": 11.5,
        "price_change_1h": 24.0,
        "oi_price_divergence": 3.1,
        "oi_acceleration": 0.8,
        "leverage_regime": "building",
        "lead_lag_score": None,
        "venue_dislocation_bps": 1.4,
        "convergence_half_life_secs": 145.0,
        "price_discovery_leader": "binance",
        "depth_recovery_speed": 0.8,
        "spread_bps": 2.3,
        "spread_zscore_4h": 0.9,
        "spread_zscore_24h": 1.2,
        "depth_imbalance_momentum": 0.05,
        "spread_regime": "normal",
        "rv_5s": 0.61,
        "rv_1m": 0.54,
        "rv_5m": 0.48,
        "vol_signature_ratio": 1.27,
        "vol_clustering": 0.36,
        "rv_regime": "elevated",
        "cum_delta_acceleration": 15.0,
        "delta_divergence": 4.0,
        "absorption_ratio": 750.0,
        "delta_regime": "aggressive_buying",
        "quality_flag": "partial_context",
    }
    store.write(DatasetName.PERPS_SIGNAL_SERIES, pl.DataFrame([row]))

    app = create_app(settings=ApiSettings(data_root=tmp_path), store=store)
    client = TestClient(app)
    latest = client.get(
        "/api/v1/perps/order_flow_toxicity/latest",
        params={"asset": "BTC", "venue_scope": "all"},
    )

    assert latest.status_code == 200
    assert latest.json()["data"]["latest_quality"] == "ok"



from __future__ import annotations

from datetime import datetime, timedelta, timezone

import polars as pl

from uforia.config import AppConfig, get_config
from uforia.ingest.deribit.client import DeribitPublicClient


def fetch_dvol(
    start: datetime,
    end: datetime,
    underlying: str = "BTC",
    client: DeribitPublicClient | None = None,
    config: AppConfig | None = None,
) -> pl.DataFrame:
    cfg = config or get_config()
    deribit = client or DeribitPublicClient(cfg)
    normalized_start = start.astimezone(timezone.utc).replace(minute=0, second=0, microsecond=0)
    normalized_end = end.astimezone(timezone.utc).replace(minute=0, second=0, microsecond=0)
    if normalized_end < normalized_start:
        raise ValueError("end must be greater than or equal to start")

    chunks: list[pl.DataFrame] = []
    cursor = normalized_end
    while cursor >= normalized_start:
        window_start = max(normalized_start, cursor - timedelta(days=30))
        result = deribit.get_volatility_index_data(
            underlying, window_start, cursor, resolution="3600"
        )
        rows = result.get("data", [])
        if not rows:
            break
        chunk = pl.DataFrame(
            {
                "ts": [datetime.fromtimestamp(item[0] / 1000, tz=timezone.utc) for item in rows],
                "date": [
                    datetime.fromtimestamp(item[0] / 1000, tz=timezone.utc).date() for item in rows
                ],
                "venue": ["deribit"] * len(rows),
                "underlying": [underlying] * len(rows),
                "source": ["dvol"] * len(rows),
                # Deribit returns DVOL in index points already, e.g. 60.89 rather than 0.6089.
                "dvol_open": [float(item[1]) for item in rows],
                "dvol_high": [float(item[2]) for item in rows],
                "dvol_low": [float(item[3]) for item in rows],
                "dvol_close": [float(item[4]) for item in rows],
                "quality_flag": ["ok"] * len(rows),
            }
        )
        chunks.append(chunk)
        continuation = result.get("continuation")
        if continuation is None:
            break
        cursor = datetime.fromtimestamp(int(continuation) / 1000, tz=timezone.utc) - timedelta(
            hours=1
        )

    if not chunks:
        return pl.DataFrame()
    return pl.concat(chunks, how="vertical_relaxed").unique(subset=["ts"]).sort("ts")



from __future__ import annotations

from datetime import datetime, timedelta, timezone

import polars as pl

from uforia.analytics import (
    build_funding_context_series,
    build_gamma_exposure_series,
    build_options_adaptive_hedge_series,
    build_options_carry_series,
    build_options_dispersion_series,
    build_options_dislocation_series,
    build_options_expiry_flow_series,
    build_options_funding_skew_series,
    build_options_ml_carry_series,
    build_options_positioning_series,
    build_options_relative_value_series,
    build_options_sessionality_series,
    build_options_signal_series,
    build_options_skew_series,
    build_options_surface_snapshot,
    build_options_surface_factor_series,
    build_straddle_breakeven_series,
    build_term_structure_momentum_series,
    build_vol_cone_series,
    build_vol_regime_transition_series,
)
from uforia.storage import DATASET_SPECS, DatasetName, DatasetStore
from tests.support import seed_options_context


def _seed_sessionality_clock_fixture(store: DatasetStore) -> None:
    rows = []
    surface_rows = []
    for ts, spot in (
        (datetime(2026, 3, 6, 8, tzinfo=timezone.utc), 70_000.0),
        (datetime(2026, 3, 7, 8, tzinfo=timezone.utc), 70_500.0),
    ):
        rows.append(
            {
                "ts": ts,
                "date": ts.date(),
                "venue": "deribit",
                "underlying": "BTC",
                "expiry": ts + timedelta(days=7),
                "instrument_name": f"BTC-{ts.isoformat()}-ATM-call",
                "option_type": "call",
                "strike": spot,
                "bid": 0.024,
                "ask": 0.026,
                "mark": 0.025,
                "mid": 0.025,
                "mark_iv": 0.55,
                "delta": 0.5,
                "gamma": 0.02,
                "vega": 0.1,
                "index_price": spot,
                "underlying_price": spot,
                "forward": spot,
                "interest_rate": 0.0,
                "open_interest": 100.0,
                "volume": 10.0,
                "quote_age_ms": 100,
                "quality_flag": "ok",
            }
        )
        surface_rows.append(
            {
                "ts": ts,
                "date": ts.date(),
                "underlying": "BTC",
                "source": "ssvi",
                "atm_iv_7d": 0.55,
                "atm_iv_14d": 0.56,
                "atm_iv_30d": 0.58,
                "atm_iv_60d": 0.60,
                "rr_25_7d": -0.05,
                "rr_25_30d": -0.03,
                "bf_25_7d": 0.02,
                "bf_25_30d": 0.03,
                "term_slope_7_30": -0.03,
                "term_slope_30_60": -0.02,
                "fit_quality": 0.001,
                "coverage_ratio": 0.8,
                "maturity_violations": 0,
                "fit_validity_flag": "ok",
                "quality_flag": "ok",
            }
        )
    store.write(DatasetName.OPTION_CHAIN_SNAPSHOT, pl.DataFrame(rows))
    store.write(DatasetName.SURFACE_FEATURES, pl.DataFrame(surface_rows))


def test_options_family_builders_produce_rows(tmp_path) -> None:
    store = DatasetStore(tmp_path)
    seed_options_context(store)

    surface_snapshot = build_options_surface_snapshot(
        store,
        start=datetime(2026, 3, 1, tzinfo=timezone.utc),
        end=datetime(2026, 3, 3, tzinfo=timezone.utc),
        underlying="BTC",
    )
    carry = build_options_carry_series(
        store,
        start=datetime(2026, 3, 1, tzinfo=timezone.utc),
        end=datetime(2026, 3, 3, tzinfo=timezone.utc),
        underlying="BTC",
    )
    skew = build_options_skew_series(
        store,
        start=datetime(2026, 3, 1, tzinfo=timezone.utc),
        end=datetime(2026, 3, 3, tzinfo=timezone.utc),
        underlying="BTC",
    )
    dislocation = build_options_dislocation_series(
        store,
        start=datetime(2026, 3, 1, tzinfo=timezone.utc),
        end=datetime(2026, 3, 3, tzinfo=timezone.utc),
        underlying="BTC",
    )
    positioning = build_options_positioning_series(
        store,
        start=datetime(2026, 3, 1, tzinfo=timezone.utc),
        end=datetime(2026, 3, 3, tzinfo=timezone.utc),
        underlying="BTC",
    )
    vol_cone = build_vol_cone_series(
        store,
        start=datetime(2026, 3, 1, tzinfo=timezone.utc),
        end=datetime(2026, 3, 3, tzinfo=timezone.utc),
        underlying="BTC",
    )
    term_momentum = build_term_structure_momentum_series(
        store,
        start=datetime(2026, 3, 1, tzinfo=timezone.utc),
        end=datetime(2026, 3, 3, tzinfo=timezone.utc),
        underlying="BTC",
    )
    gamma = build_gamma_exposure_series(
        store,
        start=datetime(2026, 3, 1, tzinfo=timezone.utc),
        end=datetime(2026, 3, 3, tzinfo=timezone.utc),
        underlying="BTC",
    )
    funding = build_funding_context_series(
        store,
        start=datetime(2026, 3, 1, tzinfo=timezone.utc),
        end=datetime(2026, 3, 3, tzinfo=timezone.utc),
        underlying="BTC",
    )
    breakeven = build_straddle_breakeven_series(
        store,
        start=datetime(2026, 3, 1, tzinfo=timezone.utc),
        end=datetime(2026, 3, 3, tzinfo=timezone.utc),
        underlying="BTC",
    )
    surface_factor = build_options_surface_factor_series(
        store,
        start=datetime(2026, 3, 1, tzinfo=timezone.utc),
        end=datetime(2026, 3, 3, tzinfo=timezone.utc),
        underlying="BTC",
    )
    ml_carry = build_options_ml_carry_series(
        store,
        start=datetime(2026, 3, 1, tzinfo=timezone.utc),
        end=datetime(2026, 3, 3, tzinfo=timezone.utc),
        underlying="BTC",
    )
    expiry_flow = build_options_expiry_flow_series(
        store,
        start=datetime(2026, 3, 1, tzinfo=timezone.utc),
        end=datetime(2026, 3, 3, tzinfo=timezone.utc),
        underlying="BTC",
    )
    adaptive = build_options_adaptive_hedge_series(
        store,
        start=datetime(2026, 3, 1, tzinfo=timezone.utc),
        end=datetime(2026, 3, 3, tzinfo=timezone.utc),
        underlying="BTC",
    )
    sessionality = build_options_sessionality_series(
        store,
        start=datetime(2026, 3, 1, tzinfo=timezone.utc),
        end=datetime(2026, 3, 3, tzinfo=timezone.utc),
        underlying="BTC",
    )
    funding_skew = build_options_funding_skew_series(
        store,
        start=datetime(2026, 3, 1, tzinfo=timezone.utc),
        end=datetime(2026, 3, 3, tzinfo=timezone.utc),
        underlying="BTC",
    )

    assert surface_snapshot.height > 0
    assert carry.height > 0
    assert skew.height > 0
    assert dislocation.height > 0
    assert positioning.height > 0
    assert vol_cone.height > 0
    assert term_momentum.height > 0
    assert gamma.height > 0
    assert funding.height > 0
    assert breakeven.height > 0
    assert surface_factor.height > 0
    assert ml_carry.height > 0
    assert expiry_flow.height > 0
    assert adaptive.height > 0
    assert sessionality.height > 0
    assert funding_skew.height > 0
    assert adaptive["route_preference"].null_count() < adaptive.height
    assert adaptive["route_score"].null_count() < adaptive.height

    store.write(DatasetName.OPTIONS_CARRY_SERIES, carry)
    store.write(
        DatasetName.OPTIONS_CARRY_SERIES,
        build_options_carry_series(
            store,
            start=datetime(2026, 3, 1, tzinfo=timezone.utc),
            end=datetime(2026, 3, 3, tzinfo=timezone.utc),
            underlying="ETH",
        ),
    )
    store.write(DatasetName.OPTIONS_SKEW_SERIES, skew)
    store.write(
        DatasetName.OPTIONS_SKEW_SERIES,
        build_options_skew_series(
            store,
            start=datetime(2026, 3, 1, tzinfo=timezone.utc),
            end=datetime(2026, 3, 3, tzinfo=timezone.utc),
            underlying="ETH",
        ),
    )
    store.write(DatasetName.OPTIONS_DISLOCATION_SERIES, dislocation)
    store.write(DatasetName.VOL_CONE_SERIES, vol_cone)
    store.write(DatasetName.TERM_STRUCTURE_MOMENTUM_SERIES, term_momentum)
    store.write(DatasetName.GAMMA_EXPOSURE_SERIES, gamma)
    store.write(DatasetName.FUNDING_CONTEXT_SERIES, funding)
    store.write(DatasetName.STRADDLE_BREAKEVEN_SERIES, breakeven)
    store.write(DatasetName.OPTIONS_SURFACE_FACTOR_SERIES, surface_factor)
    store.write(DatasetName.OPTIONS_ML_CARRY_SERIES, ml_carry)
    store.write(DatasetName.OPTIONS_EXPIRY_FLOW_SERIES, expiry_flow)
    store.write(DatasetName.OPTIONS_ADAPTIVE_HEDGE_SERIES, adaptive)
    store.write(DatasetName.OPTIONS_SESSIONALITY_SERIES, sessionality)
    store.write(DatasetName.OPTIONS_FUNDING_SKEW_SERIES, funding_skew)
    store.write(
        DatasetName.OPTIONS_DISLOCATION_SERIES,
        build_options_dislocation_series(
            store,
            start=datetime(2026, 3, 1, tzinfo=timezone.utc),
            end=datetime(2026, 3, 3, tzinfo=timezone.utc),
            underlying="ETH",
        ),
    )

    relative = build_options_relative_value_series(
        store,
        start=datetime(2026, 3, 1, tzinfo=timezone.utc),
        end=datetime(2026, 3, 3, tzinfo=timezone.utc),
    )
    store.write(DatasetName.OPTIONS_RELATIVE_VALUE_SERIES, relative)
    dispersion = build_options_dispersion_series(
        store,
        start=datetime(2026, 3, 1, tzinfo=timezone.utc),
        end=datetime(2026, 3, 3, tzinfo=timezone.utc),
    )
    store.write(DatasetName.OPTIONS_DISPERSION_SERIES, dispersion)
    signals = build_options_signal_series(
        store,
        start=datetime(2026, 3, 1, tzinfo=timezone.utc),
        end=datetime(2026, 3, 3, tzinfo=timezone.utc),
        underlying="BTC",
    )
    store.write(DatasetName.OPTIONS_SIGNAL_SERIES, signals)
    transitions = build_vol_regime_transition_series(
        store,
        start=datetime(2026, 3, 1, tzinfo=timezone.utc),
        end=datetime(2026, 3, 3, tzinfo=timezone.utc),
        underlying="BTC",
    )

    assert relative.height > 0
    assert dispersion.height > 0
    assert signals.height > 0
    assert transitions.height > 0
    assert "vol_carry_score" in signals.columns
    assert "signal_state" in signals.columns
    assert "funding_vol_context_score" in signals.columns
    assert "surface_factor_score" in signals.columns
    assert "ml_carry_toxicity_score" in signals.columns
    assert "dispersion_score" in signals.columns
    assert "vol_regime_long_threshold" in signals.columns
    assert signals["vol_regime_long_threshold"].null_count() < signals.height
    assert signals["signal_state"].null_count() < signals.height


def test_adaptive_hedge_prefers_microstructure_route_inputs(tmp_path) -> None:
    store = DatasetStore(tmp_path)
    seed_options_context(store)

    btc = build_options_adaptive_hedge_series(
        store,
        start=datetime(2026, 3, 1, tzinfo=timezone.utc),
        end=datetime(2026, 3, 3, tzinfo=timezone.utc),
        underlying="BTC",
    )
    eth = build_options_adaptive_hedge_series(
        store,
        start=datetime(2026, 3, 1, tzinfo=timezone.utc),
        end=datetime(2026, 3, 3, tzinfo=timezone.utc),
        underlying="ETH",
    )

    assert btc.filter(pl.col("route_preference") == "nearby_future").height > 0
    assert eth.filter(pl.col("route_preference") == "perp").height > 0
    assert btc["basis_mismatch_bps"].null_count() < btc.height
    assert eth["route_score"].null_count() < eth.height


def test_ml_carry_uses_forecast_rv_not_raw_carry_zscore(tmp_path) -> None:
    store = DatasetStore(tmp_path)
    seed_options_context(store)

    carry = build_options_carry_series(
        store,
        start=datetime(2026, 3, 1, tzinfo=timezone.utc),
        end=datetime(2026, 3, 3, tzinfo=timezone.utc),
        underlying="BTC",
    )
    store.write(DatasetName.OPTIONS_CARRY_SERIES, carry)

    ml_carry = build_options_ml_carry_series(
        store,
        start=datetime(2026, 3, 1, tzinfo=timezone.utc),
        end=datetime(2026, 3, 3, tzinfo=timezone.utc),
        underlying="BTC",
    )

    joined = ml_carry.join(
        carry.select(["ts", "carry_spread_7d", "realized_vol_7d"]),
        on="ts",
        how="inner",
    ).drop_nulls(["gross_carry_spread", "carry_spread_7d", "forecast_rv_7d", "realized_vol_7d"])

    assert joined.height > 0
    assert (
        joined.filter(
            (
                pl.col("gross_carry_spread")
                - (
                    pl.col("carry_spread_7d")
                    + pl.col("realized_vol_7d")
                    - pl.col("forecast_rv_7d")
                )
            ).abs()
            < 1e-9
        ).height
        > 0
    )


def test_sessionality_uses_correct_weekend_and_expiry_clock_flags(tmp_path) -> None:
    store = DatasetStore(tmp_path)
    _seed_sessionality_clock_fixture(store)

    sessionality = build_options_sessionality_series(
        store,
        start=datetime(2026, 3, 6, tzinfo=timezone.utc),
        end=datetime(2026, 3, 8, tzinfo=timezone.utc),
        underlying="BTC",
    )

    friday = sessionality.filter(pl.col("ts") == datetime(2026, 3, 6, 8, tzinfo=timezone.utc))
    saturday = sessionality.filter(pl.col("ts") == datetime(2026, 3, 7, 8, tzinfo=timezone.utc))

    assert friday.height == 1
    assert saturday.height == 1
    assert friday.item(0, "is_weekend") is False
    assert friday.item(0, "is_expiry_window") is True
    assert saturday.item(0, "is_weekend") is True
    assert saturday.item(0, "is_expiry_window") is False


def test_dispersion_uses_iv_comovement_proxy_not_iv_level_similarity(tmp_path) -> None:
    store = DatasetStore(tmp_path)
    seed_options_context(store)

    eth_surface = store.read(
        DatasetName.SURFACE_FEATURES,
        filters={"underlying": "ETH", "source": "ssvi"},
    ).with_columns(
        (pl.col("atm_iv_7d") * 1.25).alias("atm_iv_7d"),
        (pl.col("atm_iv_14d") * 1.25).alias("atm_iv_14d"),
        (pl.col("atm_iv_30d") * 1.25).alias("atm_iv_30d"),
        (pl.col("atm_iv_60d") * 1.25).alias("atm_iv_60d"),
    )
    store.write(DatasetName.SURFACE_FEATURES, eth_surface)

    dispersion = build_options_dispersion_series(
        store,
        start=datetime(2026, 3, 1, tzinfo=timezone.utc),
        end=datetime(2026, 3, 3, tzinfo=timezone.utc),
    )
    latest = (
        dispersion.filter(pl.col("tenor") == "7d")
        .drop_nulls(["implied_correlation"])
        .sort("ts")
        .tail(1)
    )

    assert latest.height == 1
    old_similarity = 1.0 - (
        abs(latest.item(0, "btc_atm_iv") - latest.item(0, "eth_atm_iv"))
        / (latest.item(0, "btc_atm_iv") + latest.item(0, "eth_atm_iv"))
    )
    assert abs(latest.item(0, "implied_correlation") - old_similarity) > 0.5


def test_expiry_flow_marks_missing_greeks_as_partial_context(tmp_path) -> None:
    store = DatasetStore(tmp_path)
    seed_options_context(store)

    btc_options = store.read(
        DatasetName.OPTION_CHAIN_SNAPSHOT,
        filters={"underlying": "BTC"},
    ).with_columns(
        pl.when(pl.col("option_type") == "put")
        .then(pl.lit(None, dtype=pl.Float64))
        .otherwise(pl.col("gamma"))
        .alias("gamma")
    )
    store.write(DatasetName.OPTION_CHAIN_SNAPSHOT, btc_options)

    expiry_flow = build_options_expiry_flow_series(
        store,
        start=datetime(2026, 3, 1, tzinfo=timezone.utc),
        end=datetime(2026, 3, 3, tzinfo=timezone.utc),
        underlying="BTC",
    )

    assert expiry_flow.height > 0
    assert expiry_flow.filter(pl.col("quality_flag") == "partial_context").height > 0
    assert expiry_flow.filter(pl.col("quality_flag") == "ok").height < expiry_flow.height


def test_read_surface_recovers_missing_latest_rows_from_option_chain(tmp_path) -> None:
    seed_store = DatasetStore(tmp_path / "seed")
    store = DatasetStore(tmp_path / "target")
    seed_options_context(seed_store)

    eth_options = seed_store.read(
        DatasetName.OPTION_CHAIN_SNAPSHOT,
        filters={"underlying": "ETH"},
    )
    eth_surface = (
        seed_store.read(
            DatasetName.SURFACE_FEATURES,
            filters={"underlying": "ETH", "source": "ssvi"},
        )
        .sort("ts")
        .head(70)
    )

    store.write(DatasetName.OPTION_CHAIN_SNAPSHOT, eth_options)
    store.write(DatasetName.SURFACE_FEATURES, eth_surface)

    carry = build_options_carry_series(
        store,
        start=datetime(2026, 3, 1, tzinfo=timezone.utc),
        end=datetime(2026, 3, 4, tzinfo=timezone.utc),
        underlying="ETH",
    )

    assert carry.height > 0
    assert carry.sort("ts").tail(1).item(0, "ts") == eth_options.sort("ts").tail(1).item(0, "ts")


def test_vol_regime_transitions_emit_current_rows_when_signal_state_is_missing(tmp_path) -> None:
    store = DatasetStore(tmp_path)
    spec = DATASET_SPECS[DatasetName.OPTIONS_SIGNAL_SERIES].schema
    rows: list[dict[str, object]] = []
    base = datetime(2026, 3, 1, 0, tzinfo=timezone.utc)
    for hour in range(3):
        ts = base + timedelta(hours=hour)
        row = {column: None for column in spec}
        row.update(
            {
                "ts": ts,
                "date": ts.date(),
                "underlying": "BTC",
                "source": "options_composite",
                "quality_flag": "insufficient_data",
                "signal_state": None,
            }
        )
        rows.append(row)
    store.write(
        DatasetName.OPTIONS_SIGNAL_SERIES,
        pl.DataFrame(rows, schema_overrides=spec),
    )

    transitions = build_vol_regime_transition_series(
        store,
        start=base,
        end=base + timedelta(hours=3),
        underlying="BTC",
    )

    assert transitions.height == 3
    assert transitions.sort("ts").tail(1).item(0, "ts") == base + timedelta(hours=2)
    assert transitions["quality_flag"].to_list() == ["insufficient_data"] * 3


def test_adaptive_hedge_falls_back_without_microstructure_route_inputs(tmp_path) -> None:
    store = DatasetStore(tmp_path)
    seed_options_context(store, include_microstructure=False)

    adaptive = build_options_adaptive_hedge_series(
        store,
        start=datetime(2026, 3, 1, tzinfo=timezone.utc),
        end=datetime(2026, 3, 3, tzinfo=timezone.utc),
        underlying="BTC",
    )

    assert adaptive.height > 0
    assert adaptive["route_preference"].drop_nulls().unique().to_list() == ["perp"]
    assert adaptive["basis_mismatch_bps"].drop_nulls().unique().to_list() == [15.0]



"""HTTP and streaming API for the uforia research platform."""

from __future__ import annotations

from typing import Any

__all__ = ["create_app"]


def __getattr__(name: str) -> Any:
    if name == "create_app":
        from .app import create_app

        return create_app
    raise AttributeError(name)



from __future__ import annotations

import time
from collections.abc import Callable
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from typing import Any
from uuid import uuid4

import orjson
import polars as pl

from uforia.config import AppConfig, get_config
from uforia.ingest.deribit.client import DeribitPublicClient
from uforia.ingest.deribit.options import fetch_option_chain_payload, normalize_option_chain_payload
from uforia.storage import DatasetName, DatasetStore, ReadMode, RunStatus, WriteMode


@dataclass(frozen=True)
class IngestionAttemptResult:
    logical_ts: datetime
    run_id: str
    status: RunStatus
    rows_written_raw: int
    rows_written_normalized: int
    error_message: str | None = None


def floor_hour(ts: datetime) -> datetime:
    return ts.astimezone(timezone.utc).replace(minute=0, second=0, microsecond=0)


def current_collectible_hour(now: datetime, grace_seconds: int) -> datetime:
    boundary = floor_hour(now)
    if now.astimezone(timezone.utc) >= boundary + timedelta(seconds=grace_seconds):
        return boundary
    return boundary - timedelta(hours=1)


def archive_deribit_options_once(
    store: DatasetStore,
    *,
    as_of: datetime,
    force: bool = False,
    underlying: str = "BTC",
    venue: str = "deribit",
    client: DeribitPublicClient | None = None,
    config: AppConfig | None = None,
) -> IngestionAttemptResult:
    cfg = config or get_config()
    deribit = client or DeribitPublicClient(cfg)
    started_at = deribit.now().astimezone(timezone.utc)
    logical_ts = floor_hour(as_of)
    run_id = uuid4().hex

    if not force and _has_successful_snapshot(
        store, logical_ts, venue=venue, underlying=underlying
    ):
        _write_ingestion_run(
            store,
            logical_ts=logical_ts,
            run_id=run_id,
            run_ts=started_at,
            dataset=DatasetName.OPTION_CHAIN_SNAPSHOT.value,
            venue=venue,
            underlying=underlying,
            status=RunStatus.SKIPPED_EXISTING,
            rows_written_raw=0,
            rows_written_normalized=0,
            duration_ms=0,
        )
        return IngestionAttemptResult(
            logical_ts=logical_ts,
            run_id=run_id,
            status=RunStatus.SKIPPED_EXISTING,
            rows_written_raw=0,
            rows_written_normalized=0,
        )

    try:
        collectible = current_collectible_hour(started_at, cfg.ingestion.archive_grace_seconds)
        if logical_ts == collectible:
            payload = fetch_option_chain_payload(
                underlying=underlying,
                venue=venue,
                client=deribit,
                config=cfg,
            )
            raw_rows = _write_raw_snapshot(
                store,
                payload=payload,
                logical_ts=logical_ts,
                run_id=run_id,
                run_ts=started_at,
            )
        else:
            archived_payload = _load_archived_payload(
                store,
                logical_ts=logical_ts,
                venue=venue,
                underlying=underlying,
            )
            if archived_payload is None:
                raise ValueError(
                    "Historical option snapshots are unavailable from the live API and no archived "
                    f"raw payload exists for {logical_ts.isoformat()}."
                )
            payload = archived_payload
            raw_rows = 0

        normalized = normalize_option_chain_payload(payload, logical_ts=logical_ts, config=cfg)
        store.write(
            DatasetName.OPTION_CHAIN_SNAPSHOT,
            normalized,
            logical_ts=logical_ts,
            run_id=run_id,
            run_ts=started_at,
            write_mode=WriteMode.IMMUTABLE,
            stats={"rows": normalized.height, "source": "deribit_options_archiver"},
        )
        finished_at = deribit.now().astimezone(timezone.utc)
        duration_ms = int((finished_at - started_at).total_seconds() * 1000)
        _write_ingestion_run(
            store,
            logical_ts=logical_ts,
            run_id=run_id,
            run_ts=started_at,
            dataset=DatasetName.OPTION_CHAIN_SNAPSHOT.value,
            venue=venue,
            underlying=underlying,
            status=RunStatus.SUCCESS,
            rows_written_raw=raw_rows,
            rows_written_normalized=normalized.height,
            duration_ms=duration_ms,
        )
        return IngestionAttemptResult(
            logical_ts=logical_ts,
            run_id=run_id,
            status=RunStatus.SUCCESS,
            rows_written_raw=raw_rows,
            rows_written_normalized=normalized.height,
        )
    except Exception as exc:
        finished_at = deribit.now().astimezone(timezone.utc)
        duration_ms = int((finished_at - started_at).total_seconds() * 1000)
        _write_ingestion_run(
            store,
            logical_ts=logical_ts,
            run_id=run_id,
            run_ts=started_at,
            dataset=DatasetName.OPTION_CHAIN_SNAPSHOT.value,
            venue=venue,
            underlying=underlying,
            status=RunStatus.FAILED,
            rows_written_raw=0,
            rows_written_normalized=0,
            duration_ms=duration_ms,
            error_message=str(exc),
        )
        return IngestionAttemptResult(
            logical_ts=logical_ts,
            run_id=run_id,
            status=RunStatus.FAILED,
            rows_written_raw=0,
            rows_written_normalized=0,
            error_message=str(exc),
        )


def archive_gap(
    store: DatasetStore,
    *,
    logical_ts: datetime,
    venue: str = "deribit",
    underlying: str = "BTC",
) -> IngestionAttemptResult:
    run_ts = datetime.now(tz=timezone.utc)
    run_id = uuid4().hex
    _write_ingestion_run(
        store,
        logical_ts=logical_ts,
        run_id=run_id,
        run_ts=run_ts,
        dataset=DatasetName.OPTION_CHAIN_SNAPSHOT.value,
        venue=venue,
        underlying=underlying,
        status=RunStatus.GAP,
        rows_written_raw=0,
        rows_written_normalized=0,
        duration_ms=0,
        error_message="missing_live_snapshot",
    )
    return IngestionAttemptResult(
        logical_ts=logical_ts,
        run_id=run_id,
        status=RunStatus.GAP,
        rows_written_raw=0,
        rows_written_normalized=0,
        error_message="missing_live_snapshot",
    )


def process_archiver_cycle(
    store: DatasetStore,
    *,
    now: datetime,
    poll_underlying: str = "BTC",
    venue: str = "deribit",
    force_current_hour: bool = False,
    client: DeribitPublicClient | None = None,
    config: AppConfig | None = None,
) -> list[IngestionAttemptResult]:
    cfg = config or get_config()
    collectible = current_collectible_hour(now, cfg.ingestion.archive_grace_seconds)
    last_attempted = _latest_attempted_logical_ts(
        store,
        venue=venue,
        underlying=poll_underlying,
    )
    results: list[IngestionAttemptResult] = []
    if last_attempted is None:
        results.append(
            archive_deribit_options_once(
                store,
                as_of=collectible,
                force=force_current_hour,
                underlying=poll_underlying,
                venue=venue,
                client=client,
                config=cfg,
            )
        )
        return results

    next_expected = last_attempted + timedelta(hours=1)
    while next_expected < collectible:
        results.append(
            archive_gap(
                store,
                logical_ts=next_expected,
                venue=venue,
                underlying=poll_underlying,
            )
        )
        next_expected += timedelta(hours=1)

    if next_expected == collectible:
        results.append(
            archive_deribit_options_once(
                store,
                as_of=collectible,
                force=force_current_hour,
                underlying=poll_underlying,
                venue=venue,
                client=client,
                config=cfg,
            )
        )
    return results


def run_deribit_options_daemon(
    store: DatasetStore,
    *,
    poll_seconds: int,
    grace_seconds: int,
    force_current_hour: bool = False,
    underlying: str = "BTC",
    venue: str = "deribit",
    client: DeribitPublicClient | None = None,
    config: AppConfig | None = None,
    sleep_fn: Callable[[float], None] = time.sleep,
) -> None:
    cfg = config or get_config()
    effective_cfg = cfg.model_copy(
        update={
            "ingestion": cfg.ingestion.model_copy(
                update={
                    "archive_poll_seconds": poll_seconds,
                    "archive_grace_seconds": grace_seconds,
                }
            )
        }
    )
    deribit = client or DeribitPublicClient(effective_cfg)
    forced_once = False
    while True:
        results = process_archiver_cycle(
            store,
            now=deribit.now(),
            poll_underlying=underlying,
            venue=venue,
            force_current_hour=force_current_hour and not forced_once,
            client=deribit,
            config=effective_cfg,
        )
        if results and force_current_hour and not forced_once:
            forced_once = True
        for result in results:
            print(
                f"{result.logical_ts.isoformat()} status={result.status.value} "
                f"raw={result.rows_written_raw} normalized={result.rows_written_normalized} "
                f"run_id={result.run_id}",
                flush=True,
            )
        sleep_fn(float(poll_seconds))


def rebuild_option_chain_from_archive(
    store: DatasetStore,
    *,
    start: datetime,
    end: datetime,
    underlying: str = "BTC",
    venue: str = "deribit",
    config: AppConfig | None = None,
) -> dict[str, int]:
    cfg = config or get_config()
    archived = store.read(
        DatasetName.RAW_DERIBIT_OPTION_SNAPSHOT,
        start=start,
        end=end,
        filters={"venue": venue, "underlying": underlying},
    ).sort("logical_ts")
    if archived.is_empty():
        return {"snapshots": 0, "rows": 0, "partitions": 0}

    snapshots = 0
    rows = 0
    partitions = 0
    for raw in archived.iter_rows(named=True):
        logical_ts = _coerce_dt(raw["logical_ts"])
        payload = orjson.loads(str(raw["payload"]))
        frame = normalize_option_chain_payload(payload, logical_ts=logical_ts, config=cfg)
        if frame.is_empty():
            continue
        written = store.write(
            DatasetName.OPTION_CHAIN_SNAPSHOT,
            frame,
            logical_ts=logical_ts,
            run_id=f"rebuild-{uuid4().hex}",
            run_ts=datetime.now(tz=timezone.utc),
            write_mode=WriteMode.IMMUTABLE,
            stats={
                "rows": frame.height,
                "source": "archive_rebuild",
                "underlying": underlying,
            },
        )
        snapshots += 1
        rows += frame.height
        partitions += len(written)
    return {"snapshots": snapshots, "rows": rows, "partitions": partitions}


def _write_raw_snapshot(
    store: DatasetStore,
    *,
    payload: dict[str, Any],
    logical_ts: datetime,
    run_id: str,
    run_ts: datetime,
) -> int:
    observed_at = _coerce_dt(payload["observed_at"])
    raw_frame = pl.DataFrame(
        [
            {
                "ts": logical_ts,
                "date": logical_ts.date(),
                "venue": payload["venue"],
                "underlying": payload["underlying"],
                "payload": orjson.dumps(_serialize_payload(payload)).decode(),
                "instrument_count": len(payload.get("books", [])),
                "quality_flag": "ok",
                "logical_ts": logical_ts,
                "run_id": run_id,
                "run_ts": run_ts,
                "write_mode": WriteMode.IMMUTABLE.value,
            }
        ]
    )
    store.write(
        DatasetName.RAW_DERIBIT_OPTION_SNAPSHOT,
        raw_frame,
        logical_ts=logical_ts,
        run_id=run_id,
        run_ts=run_ts,
        write_mode=WriteMode.IMMUTABLE,
        stats={
            "rows": 1,
            "observed_at": observed_at,
            "instrument_count": len(payload.get("books", [])),
        },
    )
    return 1


def _write_ingestion_run(
    store: DatasetStore,
    *,
    logical_ts: datetime,
    run_id: str,
    run_ts: datetime,
    dataset: str,
    venue: str,
    underlying: str,
    status: RunStatus,
    rows_written_raw: int,
    rows_written_normalized: int,
    duration_ms: int,
    error_message: str | None = None,
) -> None:
    frame = pl.DataFrame(
        [
            {
                "ts": logical_ts,
                "date": logical_ts.date(),
                "dataset": dataset,
                "venue": venue,
                "underlying": underlying,
                "status": status.value,
                "rows_written_raw": rows_written_raw,
                "rows_written_normalized": rows_written_normalized,
                "error_message": error_message,
                "duration_ms": duration_ms,
                "logical_ts": logical_ts,
                "run_id": run_id,
                "run_ts": run_ts,
                "write_mode": WriteMode.IMMUTABLE.value,
            }
        ]
    )
    store.write(
        DatasetName.INGESTION_RUNS,
        frame,
        logical_ts=logical_ts,
        run_id=run_id,
        run_ts=run_ts,
        status=status,
        write_mode=WriteMode.IMMUTABLE,
        stats={
            "dataset": dataset,
            "rows_written_raw": rows_written_raw,
            "rows_written_normalized": rows_written_normalized,
        },
        error_reason=error_message,
    )


def _has_successful_snapshot(
    store: DatasetStore,
    logical_ts: datetime,
    *,
    venue: str,
    underlying: str,
) -> bool:
    runs = store.read(
        DatasetName.INGESTION_RUNS,
        start=logical_ts,
        end=logical_ts,
        filters={
            "dataset": DatasetName.OPTION_CHAIN_SNAPSHOT.value,
            "venue": venue,
            "underlying": underlying,
            "logical_ts": logical_ts,
            "status": RunStatus.SUCCESS.value,
        },
        read_mode=ReadMode.ALL_RUNS,
    )
    return not runs.is_empty()


def _load_archived_payload(
    store: DatasetStore,
    *,
    logical_ts: datetime,
    venue: str,
    underlying: str,
) -> dict[str, Any] | None:
    raw = store.read(
        DatasetName.RAW_DERIBIT_OPTION_SNAPSHOT,
        start=logical_ts,
        end=logical_ts,
        filters={"venue": venue, "underlying": underlying, "logical_ts": logical_ts},
        read_mode=ReadMode.LATEST,
    )
    if raw.is_empty():
        return None
    payload = orjson.loads(str(raw.item(0, "payload")))
    payload["observed_at"] = payload["observed_at"]
    return payload


def _latest_attempted_logical_ts(
    store: DatasetStore,
    *,
    venue: str,
    underlying: str,
) -> datetime | None:
    runs = store.read(
        DatasetName.INGESTION_RUNS,
        filters={
            "dataset": DatasetName.OPTION_CHAIN_SNAPSHOT.value,
            "venue": venue,
            "underlying": underlying,
        },
        read_mode=ReadMode.ALL_RUNS,
    )
    if runs.is_empty():
        return None
    return runs.sort("logical_ts").tail(1).item(0, "logical_ts")


def _serialize_payload(payload: dict[str, Any]) -> dict[str, Any]:
    serialized = dict(payload)
    serialized["observed_at"] = _coerce_dt(payload["observed_at"]).isoformat()
    return serialized


def _coerce_dt(value: Any) -> datetime:
    if isinstance(value, datetime):
        return value.astimezone(timezone.utc)
    if isinstance(value, str):
        return datetime.fromisoformat(value.replace("Z", "+00:00")).astimezone(timezone.utc)
    raise TypeError(f"Unsupported datetime value: {type(value)!r}")



from __future__ import annotations

import json
from datetime import datetime, timezone
from pathlib import Path
from typing import Any

import polars as pl

HORIZON_SECONDS = {"30s": 30, "1m": 60, "2m": 120, "5m": 300}


def ensure_utc_timestamp(frame: pl.DataFrame, column: str) -> pl.DataFrame:
    return frame.with_columns(
        pl.col(column)
        .cast(pl.Datetime(time_zone="UTC"), strict=False)
        .dt.replace_time_zone("UTC")
        .alias(column)
    )


def parse_feature_payload(value: str | None) -> dict[str, float]:
    if not value:
        return {}
    try:
        parsed = json.loads(value)
    except json.JSONDecodeError:
        return {}
    if not isinstance(parsed, dict):
        return {}
    result: dict[str, float] = {}
    for key, item in parsed.items():
        try:
            result[str(key)] = float(item)
        except (TypeError, ValueError):
            continue
    return result


def encode_feature_payload(values: dict[str, Any]) -> str:
    compact = {
        key: float(value)
        for key, value in values.items()
        if value is not None and not isinstance(value, bool)
    }
    return json.dumps(compact, separators=(",", ":"), sort_keys=True)


def horizon_to_seconds(horizon: str) -> int:
    if horizon not in HORIZON_SECONDS:
        raise ValueError(f"Unsupported horizon: {horizon}")
    return HORIZON_SECONDS[horizon]


def safe_div(numerator: float | None, denominator: float | None) -> float | None:
    if numerator is None or denominator in (None, 0):
        return None
    return float(numerator) / float(denominator if denominator is not None else 1.0)


def utc_now() -> datetime:
    return datetime.now(tz=timezone.utc)


def candidate_data_roots(root: Path) -> list[Path]:
    return [root]
